import asyncio
from unittest.mock import patch

import pytest

from ophyd_async.core import (
    init_devices,
)
from ophyd_async.epics import adcore, adsimdetector
from ophyd_async.testing import get_mock_put, set_mock_value

TEST_DEADTIME = 0.1


@pytest.fixture
def driver(RE) -> adcore.ADBaseIO:
    with init_devices(mock=True):
        driver = adcore.ADBaseIO("DRV:", name="drv")
    return driver


@pytest.fixture
async def controller(RE, driver: adcore.ADBaseIO) -> adsimdetector.SimController:
    controller = adsimdetector.SimController(driver)
    controller.get_deadtime = lambda exposure: TEST_DEADTIME
    return controller


async def test_set_exposure_time_and_acquire_period_if_supplied_is_a_noop_if_no_exposure_supplied(  # noqa: E501
    controller: adsimdetector.SimController,
    driver: adcore.ADBaseIO,
):
    put_exposure = get_mock_put(driver.acquire_time)
    put_acquire_period = get_mock_put(driver.acquire_period)
    await controller.set_exposure_time_and_acquire_period_if_supplied(None)

    put_exposure.assert_not_called()
    put_acquire_period.assert_not_called()


@pytest.mark.parametrize(
    "exposure,expected_exposure,expected_acquire_period",
    [
        (0.0, 0.0, 0.1),
        (1.0, 1.0, 1.1),
        (1.5, 1.5, 1.6),
    ],
)
async def test_set_exposure_time_and_acquire_period_if_supplied_uses_deadtime(
    controller: adsimdetector.SimController,
    exposure: float,
    expected_exposure: float,
    expected_acquire_period: float,
):
    await controller.set_exposure_time_and_acquire_period_if_supplied(exposure)
    actual_exposure = await controller.driver.acquire_time.get_value()
    actual_acquire_period = await controller.driver.acquire_period.get_value()
    assert expected_exposure == actual_exposure
    assert expected_acquire_period == actual_acquire_period


@pytest.mark.timeout(18.0)
async def test_start_acquiring_driver_and_ensure_status_flags_immediate_failure(
    controller: adsimdetector.SimController,
):
    set_mock_value(controller.driver.detector_state, adcore.ADState.ERROR)
    acquiring = await controller.start_acquiring_driver_and_ensure_status()
    with pytest.raises(ValueError):
        await acquiring


async def test_start_acquiring_driver_and_ensure_status_fails_after_some_time(
    controller: adsimdetector.SimController,
):
    """This test ensures a failing status is captured halfway through acquisition.

    Real world application; it takes some time to start acquiring, and during that time
    the detector gets itself into a bad state.
    """
    set_mock_value(controller.driver.detector_state, adcore.ADState.IDLE)

    async def wait_then_fail():
        await asyncio.sleep(0)
        set_mock_value(controller.driver.detector_state, adcore.ADState.DISCONNECTED)

    await wait_then_fail()

    controller.frame_timeout = 0.1

    acquiring = await controller.start_acquiring_driver_and_ensure_status(
        start_timeout=0.2, state_timeout=0.2
    )

    with pytest.raises(
        ValueError, match="Final detector state Disconnected not in valid end states:"
    ):
        await acquiring


async def test_start_acquiring_driver_and_ensure_status_timing(
    controller: adsimdetector.SimController,
):
    """This test ensures the camera has time to return to a good state.

    Real world application; there is race condition wherein the
    detector has been asked to complete acquisition, but has not yet
    returned to a known good state before the status check.

    """
    set_mock_value(controller.driver.detector_state, adcore.ADState.ACQUIRE)

    acquiring = await controller.start_acquiring_driver_and_ensure_status(
        start_timeout=0.2, state_timeout=0.2
    )

    async def complete_acquire():
        """Return to idle state, but pretend the detector is slow."""
        await asyncio.sleep(0.05)
        set_mock_value(controller.driver.detector_state, adcore.ADState.IDLE)

    await asyncio.gather(acquiring, complete_acquire())


async def bad_observe_value(*args, **kwargs):
    "Stub to simulate a disconnected ``observe_value()``."
    if True:
        raise TimeoutError()
    yield None  # Make it a generator


@patch("ophyd_async.epics.adcore._core_logic.observe_value", bad_observe_value)
async def test_start_acquiring_driver_and_ensure_status_disconnected(
    controller: adsimdetector.SimController,
):
    """This test ensures the function behaves gracefully if no detector
    states are available.

    """
    acquiring = await controller.start_acquiring_driver_and_ensure_status(
        start_timeout=0.2, state_timeout=0.2
    )

    with pytest.raises(asyncio.TimeoutError) as exc:
        await acquiring
    assert (
        str(exc.value)
        == "Could not monitor detector state: mock+ca://DRV:DetectorState_RBV"
    )



from ophyd_async.epics.adcore._core_io import NDROIStatIO, NDROIStatNIO  # noqa: PLC2701


def test_roi_stats_channels_initialisation():
    num_channels = 5
    nd_roi_stats_io = NDROIStatIO("PREFIX:", num_channels)
    assert len(nd_roi_stats_io.channels) == num_channels
    for i in range(1, num_channels):
        assert isinstance(nd_roi_stats_io.channels[i], NDROIStatNIO)



import asyncio
import re
from unittest.mock import ANY, AsyncMock, MagicMock, call

import pytest

from ophyd_async.core import (
    Device,
    MockSignalBackend,
    SignalRW,
    SignalW,
    SoftSignalBackend,
    init_devices,
    soft_signal_r_and_setter,
    soft_signal_rw,
)
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw
from ophyd_async.testing import (
    callback_on_mock_put,
    get_mock_put,
    mock_puts_blocked,
    set_mock_put_proceeds,
    set_mock_value,
    set_mock_values,
)


async def test_mock_signal_backend():
    mock_signal = SignalRW(SoftSignalBackend(datatype=str))
    # If mock is false it will be handled like a normal signal, otherwise it will
    # initalize a new backend from the one in the line above
    await mock_signal.connect(mock=True)
    assert isinstance(mock_signal._connector.backend, MockSignalBackend)

    assert await mock_signal._connector.backend.get_value() == ""
    await mock_signal._connector.backend.put("test", True)
    assert await mock_signal._connector.backend.get_value() == "test"
    assert mock_signal._connector.backend.put_mock.call_args_list == [
        call("test", wait=True),
    ]


@pytest.mark.parametrize("epics_protocol", ["ca", "pva"])
async def test_mock_signal_backend_source(epics_protocol):
    mock_signal_rw = epics_signal_rw(
        str,
        f"{epics_protocol}://READ_PV",
        f"{epics_protocol}://WRITE_PV",
        name="mock_name",
    )
    mock_signal_r = epics_signal_r(
        str,
        f"{epics_protocol}://READ_PV",
        name="mock_name",
    )
    await mock_signal_rw.connect(mock=True)
    await mock_signal_r.connect(mock=True)

    assert mock_signal_rw.source == f"mock+{epics_protocol}://READ_PV"
    assert mock_signal_r.source == f"mock+{epics_protocol}://READ_PV"


async def test_set_mock_value():
    mock_signal = SignalRW(SoftSignalBackend(int))
    await mock_signal.connect(mock=True)
    assert await mock_signal.get_value() == 0
    assert mock_signal._connector.backend
    assert await mock_signal._connector.backend.get_value() == 0
    set_mock_value(mock_signal, 10)
    assert await mock_signal.get_value() == 10
    assert await mock_signal._connector.backend.get_value() == 10


async def test_set_mock_put_proceeds():
    mock_signal = SignalW(SoftSignalBackend(str))
    await mock_signal.connect(mock=True)

    assert isinstance(mock_signal._connector.backend, MockSignalBackend)

    assert mock_signal._connector.backend.put_proceeds.is_set() is True

    set_mock_put_proceeds(mock_signal, False)
    assert mock_signal._connector.backend.put_proceeds.is_set() is False
    set_mock_put_proceeds(mock_signal, True)
    assert mock_signal._connector.backend.put_proceeds.is_set() is True


async def test_set_mock_put_proceeds_timeout():
    mock_signal = SignalRW(SoftSignalBackend(str))
    await mock_signal.connect(mock=True)

    set_mock_put_proceeds(mock_signal, False)

    with pytest.raises(asyncio.TimeoutError):
        await mock_signal.set("test", wait=True, timeout=0.1)


async def test_put_proceeds_timeout():
    mock_signal = SignalW(SoftSignalBackend(str))
    await mock_signal.connect(mock=True)
    assert isinstance(mock_signal._connector.backend, MockSignalBackend)

    assert mock_signal._connector.backend.put_proceeds.is_set() is True

    set_mock_put_proceeds(mock_signal, False)
    assert mock_signal._connector.backend.put_proceeds.is_set() is False
    set_mock_put_proceeds(mock_signal, True)
    assert mock_signal._connector.backend.put_proceeds.is_set() is True


async def test_mock_utils_throw_error_if_backend_isnt_mock_signal_backend():
    signal = SignalRW(SoftSignalBackend(int))

    exc_msgs = []
    with pytest.raises(AssertionError) as exc:
        set_mock_value(signal, 10)
    exc_msgs.append(str(exc.value))
    with pytest.raises(AssertionError) as exc:
        get_mock_put(signal).assert_called_once_with(10)
    exc_msgs.append(str(exc.value))
    with pytest.raises(AssertionError) as exc:
        with mock_puts_blocked(signal):
            ...
    exc_msgs.append(str(exc.value))
    with pytest.raises(AssertionError) as exc:
        with callback_on_mock_put(signal, lambda x: _):
            ...
    exc_msgs.append(str(exc.value))
    with pytest.raises(AssertionError) as exc:
        set_mock_put_proceeds(signal, False)
    exc_msgs.append(str(exc.value))
    with pytest.raises(AssertionError) as exc:
        for _ in set_mock_values(signal, [10]):
            ...
    exc_msgs.append(str(exc.value))

    for msg in exc_msgs:
        assert re.match(
            r"Signal <ophyd_async.core._signal.SignalRW object at [0-9a-zA-Z]+> "
            r"not connected in mock mode",
            msg,
        )


async def test_get_mock_put():
    mock_signal = epics_signal_rw(str, "READ_PV", "WRITE_PV", name="mock_name")
    await mock_signal.connect(mock=True)
    await mock_signal.set("test_value", wait=True)

    mock = get_mock_put(mock_signal)
    mock.assert_called_once_with("test_value", wait=True)

    def err_text(text, wait):
        return (
            f"Expected: put('{re.escape(str(text))}', wait={re.escape(str(wait))})",
            "Actual: put('test_value', wait=True)",
        )

    for text, wait in [
        ("wrong_name", True),  # name wrong
        ("test_value", False),  # wait wrong
        ("wrong_name", False),  # name and wait wrong
    ]:
        with pytest.raises(AssertionError) as exc:
            mock.assert_called_once_with(text, wait=wait)
        for err_substr in err_text(text, wait):
            assert err_substr in str(exc.value)


@pytest.fixture
async def mock_signals():
    async with init_devices(mock=True):
        signal1 = epics_signal_rw(str, "READ_PV1", "WRITE_PV1", name="mock_name1")
        signal2 = epics_signal_rw(str, "READ_PV2", "WRITE_PV2", name="mock_name2")

    await signal1.set("first_value", wait=True, timeout=1)
    await signal2.set("first_value", wait=True, timeout=1)
    assert await signal1.get_value() == "first_value"
    assert await signal2.get_value() == "first_value"
    return signal1, signal2


async def test_blocks_during_put(mock_signals):
    signal1, signal2 = mock_signals

    with mock_puts_blocked(signal1, signal2):
        status1 = signal1.set("second_value", wait=True, timeout=None)
        status2 = signal2.set("second_value", wait=True, timeout=None)
        await asyncio.sleep(0.1)
        assert await signal1.get_value() == "second_value"
        assert await signal2.get_value() == "second_value"
        assert not status1.done
        assert not status2.done

    await asyncio.sleep(0.1)

    assert status1.done
    assert status2.done
    assert await signal1._connector.backend.get_value() == "second_value"
    assert await signal2._connector.backend.get_value() == "second_value"


async def test_callback_on_mock_put_as_context_manager(mock_signals):
    signal1_callbacks = MagicMock()
    signal2_callbacks = MagicMock()
    signal1, signal2 = mock_signals
    with callback_on_mock_put(signal1, signal1_callbacks):
        await signal1.set("second_value", wait=True)
    with callback_on_mock_put(signal2, signal2_callbacks):
        await signal2.set("second_value", wait=True)

    signal1_callbacks.assert_called_once_with("second_value", wait=True)
    signal2_callbacks.assert_called_once_with("second_value", wait=True)


async def test_callback_on_mock_put_not_as_context_manager():
    mock_signal = SignalRW(SoftSignalBackend(float))
    await mock_signal.connect(mock=True)
    calls = []
    callback_on_mock_put(
        mock_signal, lambda *args, **kwargs: calls.append({**kwargs, "_args": args})
    )
    await mock_signal.set(10.0)
    assert calls == [
        {
            "_args": (10.0,),
            "wait": True,
        }
    ]


async def test_async_callback_on_mock_put(mock_signals):
    signal1_callbacks = AsyncMock()
    signal2_callbacks = AsyncMock()
    signal1, signal2 = mock_signals
    with callback_on_mock_put(signal1, signal1_callbacks):
        await signal1.set("second_value", wait=True)
    with callback_on_mock_put(signal2, signal2_callbacks):
        await signal2.set("second_value", wait=True)

    signal1_callbacks.assert_awaited_once_with("second_value", wait=True)
    signal2_callbacks.assert_awaited_once_with("second_value", wait=True)


async def test_callback_on_mock_put_fails_if_args_are_not_correct():
    mock_signal = SignalRW(SoftSignalBackend(float))
    await mock_signal.connect(mock=True)

    def some_function_without_kwargs(arg):
        pass

    callback_on_mock_put(mock_signal, some_function_without_kwargs)
    with pytest.raises(TypeError) as exc:
        await mock_signal.set(10.0)
    assert str(exc.value).endswith(
        "some_function_without_kwargs() got an unexpected keyword argument 'wait'"
    )


async def test_set_mock_values(mock_signals):
    signal1, signal2 = mock_signals

    assert await signal2.get_value() == "first_value"
    for value_set in set_mock_values(signal1, ["second_value", "third_value"]):
        assert await signal1.get_value() == value_set

    iterator = set_mock_values(signal2, ["second_value", "third_value"])
    assert await signal2.get_value() == "first_value"
    next(iterator)
    assert await signal2.get_value() == "second_value"
    next(iterator)
    assert await signal2.get_value() == "third_value"


async def test_set_mock_values_exhausted_passes(mock_signals):
    signal1, signal2 = mock_signals
    for value_set in set_mock_values(
        signal1, ["second_value", "third_value"], require_all_consumed=True
    ):
        assert await signal1.get_value() == value_set

    iterator = set_mock_values(
        signal2,
        ["second_value", "third_value"] * 3,
        require_all_consumed=False,
    )
    calls = 0
    for calls, value_set in enumerate(iterator, start=1):  # noqa: B007
        assert await signal2.get_value() == value_set

    assert calls == 6


async def test_set_mock_values_exhausted_fails(mock_signals):
    signal1, _ = mock_signals

    for value_set in (
        iterator := set_mock_values(
            signal1, ["second_value", "third_value"], require_all_consumed=True
        )
    ):
        assert await signal1.get_value() == value_set
        break

    with pytest.raises(AssertionError):
        iterator.__del__()

    # Set so it doesn't raise the same error on teardown
    iterator.require_all_consumed = False


async def test_reset_mock_put_calls(mock_signals):
    signal1, _ = mock_signals
    await signal1.set("test_value", wait=True, timeout=1)
    get_mock_put(signal1).assert_called_with("test_value", wait=ANY)
    get_mock_put(signal1).reset_mock()
    with pytest.raises(AssertionError) as exc:
        get_mock_put(signal1).assert_called_with("test_value", wait=ANY)
    # Replacing spaces because they change between runners
    # (e.g the github actions runner has more)
    assert str(exc.value).replace(" ", "").replace("\n", "") == (
        "expectedcallnotfound.Expected:put('test_value',wait=<ANY>)Actual:notcalled."
    )


async def test_mock_signal_of_soft_signal_backend_receives_intial_value():
    class SomeDevice(Device):
        def __init__(self, name):
            self.my_signal = soft_signal_rw(
                datatype=int,
                initial_value=10,
            )
            super().__init__(name)

    mocked_device = SomeDevice("mocked_device")
    await mocked_device.connect(mock=True)
    soft_device = SomeDevice("soft_device")
    await soft_device.connect(mock=False)

    assert await mocked_device.my_signal.get_value() == 10
    assert await soft_device.my_signal.get_value() == 10


async def test_mock_signal_of_soft_signal_backend_receives_metadata():
    class SomeDevice(Device):
        def __init__(self, name):
            self.my_signal = soft_signal_rw(
                datatype=float, initial_value=1.0, units="mm", precision=2
            )
            super().__init__(name)

    mocked_device = SomeDevice("mocked_device")
    await mocked_device.connect(mock=True)
    soft_device = SomeDevice("soft_device")
    await soft_device.connect(mock=False)

    assert await mocked_device.my_signal.describe() == {
        "mocked_device-my_signal": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
            "source": "mock+soft://mocked_device-my_signal",
            "units": "mm",
            "precision": 2,
        }
    }
    assert await soft_device.my_signal.describe() == {
        "soft_device-my_signal": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
            "source": "soft://soft_device-my_signal",
            "units": "mm",
            "precision": 2,
        }
    }


async def test_writing_to_soft_signals_in_mock():
    class MyDevice(Device):
        def __init__(self, prefix: str, name: str = ""):
            self.signal, self.backend_put = soft_signal_r_and_setter(int)
            super().__init__(name)

        async def set(self):
            self.backend_put(1)

    device = MyDevice("-SOME-PREFIX", name="my_device")
    await device.connect(mock=True)
    assert await device.signal.get_value() == 0
    await device.set()
    assert await device.signal.get_value() == 1

    signal, backend_put = soft_signal_r_and_setter(int)
    await signal.connect(mock=False)
    assert await signal.get_value() == 0
    backend_put(100)
    assert await signal.get_value() == 100


async def test_when_put_mock_called_with_typo_then_fails_but_calling_directly_passes():
    mock_signal = SignalRW(SoftSignalBackend(int))
    await mock_signal.connect(mock=True)
    assert isinstance(mock_signal._connector.backend, MockSignalBackend)
    mock = mock_signal._connector.backend.put_mock
    with pytest.raises(AttributeError):
        mock.asssert_called_once()  # Note typo here is deliberate!
    await mock()


async def test_when_callback_on_mock_put_returns_a_value_the_readback_is_set():
    mock_signal = SignalRW(SoftSignalBackend(int))
    await mock_signal.connect(mock=True)

    callback_on_mock_put(mock_signal, lambda *_, **__: 2)

    await mock_signal.set(5)
    assert (await mock_signal.get_value()) == 2


async def test_when_callback_on_mock_put_returns_none_then_readback_is_the_setpoint():
    mock_signal = SignalRW(SoftSignalBackend(int))
    await mock_signal.connect(mock=True)

    callback_on_mock_put(mock_signal, lambda *_, **__: None)

    await mock_signal.set(5)
    assert (await mock_signal.get_value()) == 5


async def test_when_async_callback_returns_a_value_the_readback_is_set():
    mock_signal = SignalRW(SoftSignalBackend(int))
    await mock_signal.connect(mock=True)

    async def async_callback(*_, **__):
        return 2

    callback_on_mock_put(mock_signal, async_callback)

    await mock_signal.set(5)
    assert (await mock_signal.get_value()) == 2


async def test_when_async_callback_returns_none_then_readback_is_the_setpoint():
    mock_signal = SignalRW(SoftSignalBackend(int))
    await mock_signal.connect(mock=True)

    async def async_callback(*_, **__):
        return None

    callback_on_mock_put(mock_signal, async_callback)

    await mock_signal.set(5)
    assert (await mock_signal.get_value()) == 5



from ophyd_async.epics import adcore


class SimDriverIO(adcore.ADBaseIO):
    """Base class for driving simulated Areadetector IO.

    This mirrors the interface provided by ADSimDetector/db/simDetector.template.
    """

    pass



import asyncio

from ophyd_async.core import DetectorTrigger, OnOff, TriggerInfo
from ophyd_async.epics import adcore

from ._vimba_io import VimbaDriverIO, VimbaExposeOutMode, VimbaTriggerSource

TRIGGER_MODE = {
    DetectorTrigger.INTERNAL: OnOff.OFF,
    DetectorTrigger.CONSTANT_GATE: OnOff.ON,
    DetectorTrigger.VARIABLE_GATE: OnOff.ON,
    DetectorTrigger.EDGE_TRIGGER: OnOff.ON,
}

EXPOSE_OUT_MODE = {
    DetectorTrigger.INTERNAL: VimbaExposeOutMode.TIMED,
    DetectorTrigger.CONSTANT_GATE: VimbaExposeOutMode.TRIGGER_WIDTH,
    DetectorTrigger.VARIABLE_GATE: VimbaExposeOutMode.TRIGGER_WIDTH,
    DetectorTrigger.EDGE_TRIGGER: VimbaExposeOutMode.TIMED,
}


class VimbaController(adcore.ADBaseController[VimbaDriverIO]):
    """Controller for the Vimba detector."""

    def __init__(
        self,
        driver: VimbaDriverIO,
        good_states: frozenset[adcore.ADState] = adcore.DEFAULT_GOOD_STATES,
    ) -> None:
        super().__init__(driver, good_states=good_states)

    def get_deadtime(self, exposure: float | None) -> float:
        return 0.001

    async def prepare(self, trigger_info: TriggerInfo):
        await asyncio.gather(
            self.driver.trigger_mode.set(TRIGGER_MODE[trigger_info.trigger]),
            self.driver.exposure_mode.set(EXPOSE_OUT_MODE[trigger_info.trigger]),
            self.driver.num_images.set(trigger_info.total_number_of_exposures),
            self.driver.image_mode.set(adcore.ADImageMode.MULTIPLE),
        )
        if trigger_info.livetime is not None and trigger_info.trigger not in [
            DetectorTrigger.VARIABLE_GATE,
            DetectorTrigger.CONSTANT_GATE,
        ]:
            await self.driver.acquire_time.set(trigger_info.livetime)
        if trigger_info.trigger != DetectorTrigger.INTERNAL:
            self.driver.trigger_source.set(VimbaTriggerSource.LINE1)
        else:
            self.driver.trigger_source.set(VimbaTriggerSource.FREERUN)



from typing import Annotated as A

from ophyd_async.core import OnOff, SignalRW, SubsetEnum
from ophyd_async.epics import adcore
from ophyd_async.epics.core import PvSuffix


class AravisTriggerSource(SubsetEnum):
    """Which trigger source to use when TriggerMode=On."""

    LINE1 = "Line1"


class AravisDriverIO(adcore.ADBaseIO):
    """Generic Driver supporting all GiGE cameras.

    This mirrors the interface provided by ADAravis/db/aravisCamera.template.
    """

    trigger_mode: A[SignalRW[OnOff], PvSuffix.rbv("TriggerMode")]
    trigger_source: A[SignalRW[AravisTriggerSource], PvSuffix.rbv("TriggerSource")]



from ._eiger import EigerDetector
from ._eiger_controller import EigerController
from ._eiger_io import EigerDetectorIO, EigerDriverIO, EigerMonitorIO, EigerStreamIO

__all__ = [
    "EigerDetector",
    "EigerController",
    "EigerDriverIO",
    "EigerDetectorIO",
    "EigerMonitorIO",
    "EigerStreamIO",
]



import bluesky.plan_stubs as bps
import bluesky.plans as bp
import pytest
from bluesky.run_engine import RunEngine

import ophyd_async.plan_stubs as ops
from ophyd_async.core import init_devices
from ophyd_async.epics import adcore


@pytest.fixture
async def single_trigger_det_with_stats():
    async with init_devices(mock=True):
        stats = adcore.NDPluginStatsIO("PREFIX:STATS", name="stats")
        det = adcore.SingleTriggerDetector(
            drv=adcore.ADBaseIO("PREFIX:DRV"),
            plugins={"stats": stats},
            read_uncached=[stats.unique_id],
            name="det",
        )

    # Set non-default values to check they are set back
    # These are using set_mock_value to simulate the backend IOC being setup
    # in a particular way, rather than values being set by the Ophyd signals
    return det


async def test_single_trigger_det(
    single_trigger_det_with_stats: adcore.SingleTriggerDetector,
    RE: RunEngine,
):
    names = []
    docs = []
    RE.subscribe(lambda name, _: names.append(name))
    RE.subscribe(lambda _, doc: docs.append(doc))

    def plan():
        yield from ops.ensure_connected(single_trigger_det_with_stats, mock=True)
        yield from bps.abs_set(single_trigger_det_with_stats.drv.acquire_time, 0.5)
        yield from bps.abs_set(single_trigger_det_with_stats.drv.array_counter, 1)
        yield from bps.abs_set(
            single_trigger_det_with_stats.drv.image_mode, adcore.ADImageMode.CONTINUOUS
        )
        # set_mock_value(stats.unique_id, 3)
        yield from bp.count([single_trigger_det_with_stats])

    RE(plan())

    drv = single_trigger_det_with_stats.drv
    assert 1 == await drv.acquire.get_value()
    assert adcore.ADImageMode.SINGLE == await drv.image_mode.get_value()
    assert True is await drv.wait_for_plugins.get_value()

    assert names == ["start", "descriptor", "event", "stop"]
    _, descriptor, event, _ = docs
    print(descriptor)
    print(event)
    assert descriptor["configuration"]["det"]["data"]["det-drv-acquire_time"] == 0.5
    assert event["data"]["det-drv-array_counter"] == 1
    assert event["data"]["det-stats-unique_id"] == 0



import asyncio
from asyncio import Event
from pathlib import Path
from unittest.mock import ANY, AsyncMock, MagicMock

import pytest

from ophyd_async.core import init_devices
from ophyd_async.epics.odin import Odin, OdinWriter, Writing
from ophyd_async.testing import (
    callback_on_mock_put,
    get_mock_put,
    set_mock_value,
)

ODIN_DETECTOR_NAME = "odin_detector"
EIGER_BIT_DEPTH = 16

OdinDriverAndWriter = tuple[Odin, OdinWriter]


@pytest.fixture
def odin_driver_and_writer(RE) -> OdinDriverAndWriter:
    eiger_bit_depth = AsyncMock(get_value=AsyncMock(return_value=EIGER_BIT_DEPTH))
    with init_devices(mock=True):
        driver = Odin("")
        writer = OdinWriter(MagicMock(), driver, eiger_bit_depth)
    writer._path_provider.return_value.filename = "filename.h5"  # type: ignore
    return driver, writer


def initialise_signals_to_armed(driver):
    set_mock_value(driver.meta_active, "Active")
    set_mock_value(driver.capture_rbv, "Capturing")
    set_mock_value(driver.meta_writing, "Writing")
    set_mock_value(driver.meta_file_name, "filename.h5")
    set_mock_value(driver.id, "filename.h5")


async def test_when_open_called_then_file_correctly_set(
    odin_driver_and_writer: OdinDriverAndWriter, tmp_path: Path
):
    driver, writer = odin_driver_and_writer
    initialise_signals_to_armed(driver)
    path_info = writer._path_provider.return_value  # type: ignore
    path_info.directory_path = tmp_path
    expected_filename = "filename.h5"

    await writer.open(ODIN_DETECTOR_NAME)

    get_mock_put(driver.file_path).assert_called_once_with(str(tmp_path), wait=ANY)
    get_mock_put(driver.file_name).assert_called_once_with(expected_filename, wait=ANY)


async def test_when_open_called_then_all_expected_signals_set(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    initialise_signals_to_armed(driver)

    await writer.open(ODIN_DETECTOR_NAME)

    get_mock_put(driver.data_type).assert_called_once_with("UInt16", wait=ANY)
    get_mock_put(driver.num_to_capture).assert_called_once_with(0, wait=ANY)

    get_mock_put(driver.capture).assert_called_once_with(Writing.CAPTURE, wait=ANY)


async def test_bit_depth_is_passed_before_open_and_set_to_data_type_after_open(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    initialise_signals_to_armed(driver)

    assert await writer._detector_bit_depth().get_value() == EIGER_BIT_DEPTH
    assert await driver.data_type.get_value() == ""
    await writer.open(ODIN_DETECTOR_NAME)
    get_mock_put(driver.data_type).assert_called_once_with(
        f"UInt{EIGER_BIT_DEPTH}", wait=ANY
    )


async def test_given_data_shape_set_when_open_called_then_describe_has_correct_shape(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    initialise_signals_to_armed(driver)

    set_mock_value(driver.image_width, 1024)
    set_mock_value(driver.image_height, 768)
    description = await writer.open(ODIN_DETECTOR_NAME)
    assert description["data"]["shape"] == [1, 768, 1024]


async def test_when_closed_then_data_capture_turned_off(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    await writer.close()
    get_mock_put(driver.capture).assert_called_once_with(Writing.DONE, wait=ANY)


@pytest.mark.asyncio
async def test_wait_for_active_and_file_names_before_capture_then_wait_for_writing(
    odin_driver_and_writer,
):
    driver, writer = odin_driver_and_writer

    file_name_is_set = Event()
    capture_is_set = Event()
    callback_on_mock_put(
        driver.file_name, lambda *args, **kwargs: file_name_is_set.set()
    )
    callback_on_mock_put(driver.capture, lambda *args, **kwargs: capture_is_set.set())

    async def set_waited_signals():
        set_mock_value(driver.meta_active, "Active")
        set_mock_value(driver.id, "filename.h5")
        set_mock_value(driver.meta_file_name, "filename.h5")

    async def set_ready_signals():
        set_mock_value(driver.meta_writing, "Writing")
        set_mock_value(driver.capture_rbv, "Capturing")

    async def wait_and_set_signals():
        # Block until filename is set
        await file_name_is_set.wait()
        # Allow writer.open to proceed to wait_for_value.
        await asyncio.sleep(0.1)
        # writer.open now waits on signals; set these, and unset event
        await set_waited_signals()
        # Block until capture sets event
        await capture_is_set.wait()
        # Allow writer.open to proceed to wait_for_value.
        await asyncio.sleep(0.1)
        # writer.open now waits on signals; set these
        await set_ready_signals()

    await asyncio.gather(writer.open(ODIN_DETECTOR_NAME), wait_and_set_signals())



from typing import Annotated as A

from ophyd_async.core import OnOff, SignalRW, StrictEnum
from ophyd_async.epics import adcore
from ophyd_async.epics.core import PvSuffix


class VimbaConvertFormat(StrictEnum):
    """Convert pixel format for the Vimba detector."""

    NONE = "None"
    MONO8 = "Mono8"
    MONO16 = "Mono16"
    RGB8 = "RGB8"
    RGB16 = "RGB16"


class VimbaTriggerSource(StrictEnum):
    """Mode for the source of triggers on the Vimbda."""

    FREERUN = "Freerun"
    LINE1 = "Line1"
    LINE2 = "Line2"
    FIXED_RATE = "FixedRate"
    SOFTWARE = "Software"
    ACTION0 = "Action0"
    ACTION1 = "Action1"


class VimbaOverlap(StrictEnum):
    """Overlap modes for the Vimba detector."""

    OFF = OnOff.OFF.value
    PREV_FRAME = "PreviousFrame"


class VimbaExposeOutMode(StrictEnum):
    """Exposure control modes for Vimba detectors."""

    TIMED = "Timed"  # Use ExposureTime PV
    TRIGGER_WIDTH = "TriggerWidth"  # Expose for length of high signal


class VimbaDriverIO(adcore.ADBaseIO):
    """Mirrors the interface provided by ADVimba/db/vimba.template."""

    convert_pixel_format: A[
        SignalRW[VimbaConvertFormat], PvSuffix("ConvertPixelFormat")
    ]
    trigger_source: A[SignalRW[VimbaTriggerSource], PvSuffix("TriggerSource")]
    trigger_mode: A[SignalRW[OnOff], PvSuffix("TriggerMode")]
    trigger_overlap: A[SignalRW[VimbaOverlap], PvSuffix("TriggerOverlap")]
    exposure_mode: A[SignalRW[VimbaExposeOutMode], PvSuffix("ExposureMode")]



import asyncio
import re
from unittest.mock import MagicMock, call, patch

import numpy as np
import pytest
from bluesky.protocols import Reading, Subscribable

from ophyd_async.core import (
    Array1D,
    Callback,
    Signal,
    SignalBackend,
    SignalDatatype,
    derived_signal_r,
    derived_signal_rw,
    derived_signal_w,
    soft_signal_r_and_setter,
    soft_signal_rw,
)
from ophyd_async.testing import (
    BeamstopPosition,
    Exploder,
    MovableBeamstop,
    ReadOnlyBeamstop,
    assert_describe_signal,
    assert_reading,
    assert_value,
    get_mock,
    partial_reading,
)


@pytest.fixture
def movable_beamstop() -> MovableBeamstop:
    return MovableBeamstop("device")


@pytest.fixture
def readonly_beamstop() -> ReadOnlyBeamstop:
    return ReadOnlyBeamstop("device")


def _get_position(foo: float, bar: float) -> BeamstopPosition:
    if abs(foo) < 1 and abs(bar) < 2:
        return BeamstopPosition.IN_POSITION
    else:
        return BeamstopPosition.OUT_OF_POSITION


def _get_position_wrong_args(x: float, y: float) -> BeamstopPosition:
    if abs(x) < 1 and abs(y) < 2:
        return BeamstopPosition.IN_POSITION
    else:
        return BeamstopPosition.OUT_OF_POSITION


@pytest.mark.parametrize(
    "x, y, position",
    [
        (0, 0, BeamstopPosition.IN_POSITION),
        (3, 5, BeamstopPosition.OUT_OF_POSITION),
    ],
)
@pytest.mark.parametrize("cls", [ReadOnlyBeamstop, MovableBeamstop])
async def test_get_returns_right_position(
    cls: type[ReadOnlyBeamstop | MovableBeamstop],
    x: float,
    y: float,
    position: BeamstopPosition,
):
    inst = cls("inst")
    await inst.x.set(x)
    await inst.y.set(y)
    await assert_value(inst.position, position)
    await assert_reading(inst.position, {"inst-position": partial_reading(position)})
    await assert_describe_signal(
        inst.position,
        choices=[
            "In position",
            "Out of position",
        ],
        dtype="string",
        dtype_numpy="|S40",
        shape=[],
    )


@pytest.mark.parametrize("cls", [ReadOnlyBeamstop, MovableBeamstop])
async def test_monitoring_position(cls: type[ReadOnlyBeamstop | MovableBeamstop]):
    results = asyncio.Queue[BeamstopPosition]()
    inst = cls("inst")
    inst.position.subscribe_reading(results.put_nowait)
    assert (await results.get())["inst-position"][
        "value"
    ] == BeamstopPosition.IN_POSITION
    assert results.empty()
    await inst.x.set(3)
    assert (await results.get())["inst-position"][
        "value"
    ] == BeamstopPosition.OUT_OF_POSITION
    assert results.empty()
    await inst.y.set(5)
    assert (await results.get())["inst-position"][
        "value"
    ] == BeamstopPosition.OUT_OF_POSITION
    assert results.empty()
    await asyncio.gather(inst.x.set(0), inst.y.set(0))
    assert (await results.get())["inst-position"][
        "value"
    ] == BeamstopPosition.OUT_OF_POSITION
    assert (await results.get())["inst-position"][
        "value"
    ] == BeamstopPosition.IN_POSITION
    assert results.empty()


async def test_setting_position():
    inst = MovableBeamstop("inst")
    # Connect in mock mode so we can see what would have been set
    await inst.connect(mock=True)
    m = get_mock(inst)
    await inst.position.set(BeamstopPosition.OUT_OF_POSITION)
    assert m.mock_calls == [
        call.position.put(BeamstopPosition.OUT_OF_POSITION, wait=True),
        call.x.put(3, wait=True),
        call.y.put(5, wait=True),
    ]
    m.reset_mock()
    await inst.position.set(BeamstopPosition.IN_POSITION)
    assert m.mock_calls == [
        call.position.put(BeamstopPosition.IN_POSITION, wait=True),
        call.x.put(0, wait=True),
        call.y.put(0, wait=True),
    ]


async def test_setting_all():
    inst = Exploder(3, "exploder")
    await assert_reading(
        inst, {f"exploder-signals-{i}": partial_reading(0) for i in range(1, 4)}
    )
    await inst.set_all.set(5)
    await assert_reading(
        inst, {f"exploder-signals-{i}": partial_reading(5) for i in range(1, 4)}
    )


@pytest.mark.parametrize(
    "func, expected_msg, args",
    [
        (
            _get_position_wrong_args,
            "Expected the following to be passed as keyword arguments "
            "{'x': <class 'float'>, 'y': <class 'float'>}, "
            "got {'foo': <class 'float'>, 'bar': <class 'float'>}",
            {"foo": soft_signal_rw(float), "bar": soft_signal_rw(float)},
        ),
        (
            _get_position,
            "Expected the following to be passed as keyword arguments "
            "{'foo': <class 'float'>, 'bar': <class 'float'>}, "
            "got {'foo': <class 'int'>, 'bar': <class 'int'>}",
            {
                "foo": soft_signal_rw(int),
                "bar": soft_signal_rw(int),
            },  # Signals are of wrong type.
        ),
    ],
)
def test_mismatching_args_and_types(func, expected_msg, args):
    with pytest.raises(TypeError, match=re.escape(expected_msg)):
        derived_signal_r(func, **args)


def _get(ts: int) -> float:
    return ts


async def _put(value: float) -> None:
    pass


@pytest.fixture
def derived_signal_backend() -> SignalBackend[SignalDatatype]:
    signal_rw = soft_signal_rw(int, initial_value=4)
    derived = derived_signal_rw(_get, _put, ts=signal_rw)
    return derived._connector.backend


async def test_derived_signal_rw_works_with_signal_r():
    signal_r, _ = soft_signal_r_and_setter(int, initial_value=4)
    derived = derived_signal_rw(_get, _put, ts=signal_r)
    assert await derived.get_value() == 4


async def test_derived_signal_allows_literals():
    signal_rw = soft_signal_rw(int, 0, "TEST")

    def _add_const_to_value(signal: int, const: int) -> int:
        return const + signal

    signal_r = derived_signal_r(
        _add_const_to_value,
        signal=signal_rw,
        const=24,
    )
    assert await signal_r.get_value() == 24
    await signal_rw.set(10)
    assert await signal_r.get_value() == 34


async def test_validate_by_type(derived_signal_backend: SignalBackend):
    def float_get(ts: float) -> float:
        return ts

    with pytest.raises(TypeError, match=re.escape(" is not an instance of")):
        derived_signal_rw(float_get, _put, ts=Signal(derived_signal_backend))


async def test_set_derived_not_initialized():
    sig = derived_signal_r(_get, ts=soft_signal_rw(int, initial_value=4))
    with pytest.raises(
        RuntimeError,
        match="Cannot put as no set_derived method given",
    ):
        await sig._connector.backend.put(1.0, True)


async def test_derived_update_cached_reading_not_initialized(
    derived_signal_backend: SignalBackend,
):
    class TestCls(Subscribable):
        def subscribe_reading(self, function: Callback) -> None:
            pass

        subscribe = subscribe_reading

        def clear_sub(self, function: Callback) -> None:
            function("")

        @property
        def name(self) -> str:
            return ""

    with patch.object(
        derived_signal_backend.transformer,  # type: ignore
        "raw_and_transform_subscribables",
        {"raw_device": TestCls()},
    ):
        with pytest.raises(
            RuntimeError,
            match=re.escape(
                "Cannot update cached reading as it has not been initialised"
            ),
        ):  # noqa: E501
            derived_signal_backend.set_callback(None)


async def test_set_derived_callback_already_set(derived_signal_backend: SignalBackend):
    mock_callback = MagicMock(Callback[Reading])
    derived_signal_backend.set_callback(mock_callback)
    with pytest.raises(RuntimeError, match=re.escape("Callback already set for")):
        derived_signal_backend.set_callback(mock_callback)


@patch("ophyd_async.core._derived_signal.get_type_hints", return_value={})
def test_get_return_datatype_no_type(movable_beamstop: MovableBeamstop):
    with pytest.raises(
        TypeError, match=re.escape("does not have a type hint for it's return value")
    ):
        derived_signal_r(movable_beamstop._get_position)


@patch("ophyd_async.core._derived_signal.get_type_hints", return_value={})
def test_get_first_arg_datatype_no_type(movable_beamstop: MovableBeamstop):
    with pytest.raises(
        TypeError, match=re.escape("does not have a type hinted argument")
    ):
        derived_signal_w(movable_beamstop._set_from_position)


def test_derived_signal_rw_type_error(movable_beamstop: MovableBeamstop):
    with patch.object(
        movable_beamstop, "_set_from_position", movable_beamstop._get_position
    ):  # noqa: E501
        with pytest.raises(TypeError):
            derived_signal_rw(
                movable_beamstop._get_position, movable_beamstop._set_from_position
            )  # noqa: E501


async def test_datakey_shape():
    def _total(spectrum: Array1D[np.float64]) -> float:
        return np.sum(spectrum, dtype=np.float64)

    spectrum = soft_signal_rw(Array1D[np.float64])
    intensity = derived_signal_r(_total, spectrum=spectrum)
    with pytest.raises(
        TypeError,
        match="Can't make shape for 0.0 with SignalDataType: <class 'numpy.float64'>",
    ):
        await intensity.describe()



import asyncio
from collections.abc import Awaitable, Callable
from typing import cast
from unittest.mock import patch

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    TriggerInfo,
)
from ophyd_async.epics import adcore
from ophyd_async.epics.adpilatus import (
    PilatusDetector,
    PilatusDriverIO,
    PilatusReadoutTime,
    PilatusTriggerMode,
)
from ophyd_async.testing import set_mock_value


@pytest.fixture
def test_adpilatus(ad_standard_det_factory) -> PilatusDetector:
    return ad_standard_det_factory(PilatusDetector)


@pytest.mark.parametrize(
    "readout_time",
    [PilatusReadoutTime.PILATUS2, PilatusReadoutTime.PILATUS3],
)
async def test_given_a_different_readout_time_then_deadtime_is_changed(
    ad_standard_det_factory, readout_time: float
):
    detector: PilatusDetector = ad_standard_det_factory(
        PilatusDetector, readout_time=readout_time
    )
    assert detector._controller.get_deadtime(None) == readout_time


async def test_deadtime_invariant(
    test_adpilatus: PilatusDetector,
):
    pilatus_controller = test_adpilatus._controller
    # deadtime invariant with exposure time
    assert pilatus_controller.get_deadtime(0) == 0.95e-3
    assert pilatus_controller.get_deadtime(500) == 0.95e-3


@pytest.mark.parametrize(
    "detector_trigger,expected_trigger_mode",
    [
        (DetectorTrigger.INTERNAL, PilatusTriggerMode.INTERNAL),
        (DetectorTrigger.CONSTANT_GATE, PilatusTriggerMode.EXT_ENABLE),
        (DetectorTrigger.VARIABLE_GATE, PilatusTriggerMode.EXT_ENABLE),
        (DetectorTrigger.EDGE_TRIGGER, PilatusTriggerMode.EXT_TRIGGER),
    ],
)
async def test_trigger_mode_set(
    test_adpilatus: PilatusDetector,
    detector_trigger: DetectorTrigger,
    expected_trigger_mode: PilatusTriggerMode,
):
    async def trigger_and_complete():
        set_mock_value(test_adpilatus.driver.armed, True)
        await test_adpilatus._controller.prepare(
            TriggerInfo(number_of_events=1, trigger=detector_trigger)
        )
        await test_adpilatus._controller.arm()
        await test_adpilatus._controller.wait_for_idle()

    await _trigger(test_adpilatus, expected_trigger_mode, trigger_and_complete)


async def test_trigger_mode_set_without_armed_pv(
    test_adpilatus: PilatusDetector,
):
    async def trigger_and_complete():
        await test_adpilatus._controller.prepare(
            TriggerInfo(number_of_events=1, trigger=DetectorTrigger.INTERNAL)
        )
        await test_adpilatus._controller.arm()
        await test_adpilatus._controller.wait_for_idle()

    with patch(
        "ophyd_async.epics.adpilatus._pilatus_controller.DEFAULT_TIMEOUT",
        0.1,
    ):
        with pytest.raises(asyncio.TimeoutError):
            await _trigger(
                test_adpilatus,
                PilatusTriggerMode.INTERNAL,
                trigger_and_complete,
            )


async def _trigger(
    test_adpilatus: PilatusDetector,
    expected_trigger_mode: PilatusTriggerMode,
    trigger_and_complete: Callable[[], Awaitable],
):
    pilatus_driver = test_adpilatus.driver
    # Default TriggerMode
    assert await pilatus_driver.trigger_mode.get_value() == PilatusTriggerMode.INTERNAL

    await trigger_and_complete()

    # TriggerSource changes
    assert (await pilatus_driver.trigger_mode.get_value()) == expected_trigger_mode


async def test_exposure_time_and_acquire_period_set(
    test_adpilatus: PilatusDetector,
):
    async def dummy_open(name: str, exposures_per_event: int = 1):
        test_adpilatus._writer._exposures_per_event = exposures_per_event
        return {}

    test_adpilatus._writer.open = dummy_open
    set_mock_value(test_adpilatus.driver.armed, True)
    await test_adpilatus.prepare(
        TriggerInfo(
            number_of_events=1,
            trigger=DetectorTrigger.INTERNAL,
            deadtime=1.0,
            livetime=1.0,
        )
    )
    assert (await test_adpilatus.driver.acquire_time.get_value()) == 1.0
    assert (await test_adpilatus.driver.acquire_period.get_value()) == 1.0 + 950e-6


async def test_pilatus_controller(test_adpilatus: PilatusDetector):
    pilatus = test_adpilatus._controller
    pilatus_driver = cast(PilatusDriverIO, test_adpilatus.driver)
    set_mock_value(pilatus_driver.armed, True)
    await pilatus.prepare(
        TriggerInfo(number_of_events=1, trigger=DetectorTrigger.CONSTANT_GATE)
    )
    await pilatus.arm()
    await pilatus.wait_for_idle()

    assert await pilatus_driver.num_images.get_value() == 1
    assert await pilatus_driver.image_mode.get_value() == adcore.ADImageMode.MULTIPLE
    assert (
        await pilatus_driver.trigger_mode.get_value() == PilatusTriggerMode.EXT_ENABLE
    )

    assert await pilatus_driver.acquire.get_value() is True

    await pilatus.disarm()

    assert await pilatus_driver.acquire.get_value() is False



from ._block import (
    CommonPandaBlocks,
    DataBlock,
    InencBlock,
    PandaBitMux,
    PandaCaptureMode,
    PandaPcompDirection,
    PandaPosMux,
    PandaTimeUnits,
    PcapBlock,
    PcompBlock,
    PulseBlock,
    SeqBlock,
)
from ._control import PandaPcapController
from ._hdf_panda import HDFPanda
from ._table import (
    DatasetTable,
    PandaHdf5DatasetType,
    SeqTable,
    SeqTrigger,
)
from ._trigger import (
    PcompInfo,
    PosOutScaleOffset,
    ScanSpecInfo,
    ScanSpecSeqTableTriggerLogic,
    SeqTableInfo,
    StaticPcompTriggerLogic,
    StaticSeqTableTriggerLogic,
)
from ._writer import PandaHDFWriter

__all__ = [
    "CommonPandaBlocks",
    "DataBlock",
    "InencBlock",
    "PandaBitMux",
    "PandaCaptureMode",
    "PcapBlock",
    "PcompBlock",
    "PandaPcompDirection",
    "PandaPosMux",
    "PulseBlock",
    "SeqBlock",
    "PandaTimeUnits",
    "HDFPanda",
    "PandaHDFWriter",
    "PandaPcapController",
    "DatasetTable",
    "PandaHdf5DatasetType",
    "SeqTable",
    "SeqTrigger",
    "PcompInfo",
    "SeqTableInfo",
    "StaticPcompTriggerLogic",
    "StaticSeqTableTriggerLogic",
    "ScanSpecInfo",
    "ScanSpecSeqTableTriggerLogic",
    "PosOutScaleOffset",
]



from unittest.mock import AsyncMock, MagicMock

import pytest

from ophyd_async.core import DetectorTrigger, StandardDetector, TriggerInfo
from ophyd_async.core._detector import _ensure_trigger_info_exists  # noqa: PLC2701


@pytest.fixture
def mock_controller() -> MagicMock:
    controller = MagicMock()
    controller.get_deadtime = MagicMock(return_value=0.5)
    controller.get_exposures_per_event = AsyncMock(return_value=1)
    controller.prepare = AsyncMock(return_value={})
    controller.arm = AsyncMock()
    return controller


@pytest.fixture
def mock_writer() -> MagicMock:
    writer = MagicMock()
    writer.open = AsyncMock(return_value={})
    writer.get_indices_written = AsyncMock(return_value=0)
    return writer


@pytest.fixture
def standard_detector(
    mock_controller: MagicMock, mock_writer: MagicMock
) -> StandardDetector:
    return StandardDetector(
        controller=mock_controller,
        writer=mock_writer,
        config_sigs=[],
        name="test_detector",
    )


async def test_prepare_internal_trigger(standard_detector: StandardDetector) -> None:
    trigger_info = TriggerInfo(
        number_of_events=1,
        trigger=DetectorTrigger.INTERNAL,
        deadtime=0,
        livetime=None,
        exposure_timeout=None,
    )
    await standard_detector.prepare(trigger_info)
    assert standard_detector._trigger_info == trigger_info
    assert standard_detector._number_of_events_iter is not None
    assert standard_detector._initial_frame == 0
    assert standard_detector._name == "test_detector"
    standard_detector._writer.open.assert_called_once_with(
        standard_detector._name, trigger_info.exposures_per_event or 1
    )  # type: ignore
    standard_detector._controller.prepare.assert_called_once_with(trigger_info)  # type: ignore


async def test_prepare_external_trigger(standard_detector: StandardDetector) -> None:
    trigger_info = TriggerInfo(
        number_of_events=1,
        trigger=DetectorTrigger.EDGE_TRIGGER,
        deadtime=1.0,
        livetime=None,
        exposure_timeout=None,
    )
    await standard_detector.prepare(trigger_info)
    assert standard_detector._trigger_info == trigger_info
    assert standard_detector._number_of_events_iter is not None
    assert standard_detector._initial_frame == 0
    standard_detector._writer.open.assert_called_once_with(
        standard_detector.name, trigger_info.exposures_per_event or 1
    )  # type: ignore
    standard_detector._controller.prepare.assert_called_once_with(trigger_info)  # type: ignore
    standard_detector._controller.arm.assert_called_once()  # type: ignore


async def test_prepare_external_trigger_no_deadtime(
    standard_detector: StandardDetector,
) -> None:
    trigger_info = TriggerInfo(
        number_of_events=1,
        trigger=DetectorTrigger.EDGE_TRIGGER,
        deadtime=0,  # Less than the required 0.5 set in the fixture
        livetime=None,
        exposure_timeout=None,
    )
    with pytest.raises(
        ValueError,
        match=r"Deadtime must be supplied when in externally triggered mode",
    ):
        await standard_detector.prepare(trigger_info)


def test_ensure_trigger_info_exists_success() -> None:
    trigger_info = TriggerInfo(number_of_events=1)
    assert isinstance(
        _ensure_trigger_info_exists(trigger_info=trigger_info),
        TriggerInfo,
    )


def test_ensure_trigger_info_exists_raises() -> None:
    with pytest.raises(
        RuntimeError, match="Trigger info must be set before calling this method."
    ):
        assert isinstance(
            _ensure_trigger_info_exists(trigger_info=None),
            TriggerInfo,
        )



import xml.etree.ElementTree as ET
from unittest.mock import patch

import pytest

from ophyd_async.core import (
    DatasetDescriber,
    PathProvider,
    StandardDetector,
    StaticPathProvider,
    init_devices,
)
from ophyd_async.epics import adaravis, adcore, adkinetix, adpilatus, advimba
from ophyd_async.epics.adpilatus import PilatusReadoutTime
from ophyd_async.epics.core import epics_signal_r
from ophyd_async.plan_stubs import setup_ndattributes, setup_ndstats_sum
from ophyd_async.testing import callback_on_mock_put, set_mock_value

DETECTOR_NAME = "test"


class DummyDatasetDescriber(DatasetDescriber):
    async def np_datatype(self) -> str:
        return "<u2"

    async def shape(self) -> tuple[int, int]:
        return (10, 10)


@pytest.fixture
async def hdf_writer(
    RE, static_path_provider: StaticPathProvider
) -> adcore.ADHDFWriter:
    async with init_devices(mock=True):
        hdf = adcore.NDFileHDFIO("HDF:")

    writer = adcore.ADHDFWriter(
        hdf,
        static_path_provider,
        DummyDatasetDescriber(),
        {},
    )

    def on_set_file_path_callback(value: str, wait: bool = True):
        """Mock a successful directory & file creation"""
        set_mock_value(writer.fileio.file_path_exists, True)
        set_mock_value(
            writer.fileio.full_file_name,
            f"{value}/{static_path_provider._filename_provider()}{writer._file_extension}",
        )

    callback_on_mock_put(writer.fileio.file_path, on_set_file_path_callback)

    return writer


@pytest.fixture
async def tiff_writer(
    RE, static_path_provider: StaticPathProvider
) -> adcore.ADTIFFWriter:
    async with init_devices(mock=True):
        tiff = adcore.NDFilePluginIO("TIFF:")

    writer = adcore.ADTIFFWriter(
        tiff, static_path_provider, DummyDatasetDescriber(), {}
    )

    def on_set_file_path_callback(value: str, wait: bool = True):
        """Mock a successful directory & file creation"""
        set_mock_value(writer.fileio.file_path_exists, True)
        set_mock_value(
            writer.fileio.full_file_name,
            f"{value}/{static_path_provider._filename_provider()}{writer._file_extension}",
        )

    callback_on_mock_put(writer.fileio.file_path, on_set_file_path_callback)

    return writer


@pytest.fixture
async def hdf_writer_with_stats(
    RE, static_path_provider: StaticPathProvider
) -> adcore.ADHDFWriter:
    async with init_devices(mock=True):
        hdf = adcore.NDFileHDFIO("HDF:")
        stats = adcore.NDPluginStatsIO("FOO:")

    # Set number of frames per chunk to something reasonable
    set_mock_value(hdf.num_frames_chunks, 2)

    writer = adcore.ADHDFWriter(
        hdf,
        static_path_provider,
        DummyDatasetDescriber(),
        {"stats": stats},
    )

    def on_set_file_path_callback(value: str, wait: bool = True):
        """Mock a successful directory & file creation"""
        set_mock_value(writer.fileio.file_path_exists, True)
        set_mock_value(
            writer.fileio.full_file_name,
            f"{value}/{static_path_provider._filename_provider()}{writer._file_extension}",
        )

    callback_on_mock_put(writer.fileio.file_path, on_set_file_path_callback)

    return writer


@pytest.fixture
async def detectors(
    static_path_provider: PathProvider,
) -> list[StandardDetector]:
    detectors = []
    async with init_devices(mock=True):
        detectors.append(advimba.VimbaDetector("VIMBA:", static_path_provider))
        detectors.append(adkinetix.KinetixDetector("KINETIX:", static_path_provider))
        detectors.append(
            adpilatus.PilatusDetector(
                "PILATUS:", static_path_provider, PilatusReadoutTime.PILATUS3
            )
        )
        detectors.append(adaravis.AravisDetector("ADARAVIS:", static_path_provider))
    return detectors


async def test_hdf_writer_file_not_found(hdf_writer: adcore.ADHDFWriter):
    callback_on_mock_put(hdf_writer.fileio.file_path, lambda value, wait: None)
    with pytest.raises(
        FileNotFoundError, match=r"Path .* doesn't exist or not writable!"
    ):
        await hdf_writer.open(DETECTOR_NAME)


async def test_hdf_writer_collect_stream_docs(hdf_writer: adcore.ADHDFWriter):
    assert hdf_writer._composer is None
    await hdf_writer.open("test", exposures_per_event=1)
    [item async for item in hdf_writer.collect_stream_docs(DETECTOR_NAME, 1)]

    assert hdf_writer._composer


async def test_tiff_writer_collect_stream_docs(tiff_writer: adcore.ADTIFFWriter):
    assert tiff_writer._emitted_resource is None
    await tiff_writer.open("test", exposures_per_event=1)
    [item async for item in tiff_writer.collect_stream_docs(DETECTOR_NAME, 1)]
    assert tiff_writer._emitted_resource


async def test_stats_describe_when_plugin_configured(
    hdf_writer_with_stats: adcore.ADHDFWriter,
):
    assert hdf_writer_with_stats._composer is None
    set_mock_value(hdf_writer_with_stats.fileio.file_path_exists, True)
    assert hdf_writer_with_stats._plugins is not None
    set_mock_value(
        hdf_writer_with_stats._plugins["stats"].nd_attributes_file,
        """<?xml version='1.0' encoding='utf-8'?>
<Attributes>
    <Attribute
        name="mydetector-sum"
        type="PARAM"
        source="TOTAL" addr="0"
        datatype="DOUBLE"
        description="Sum of each detector frame" />
    <Attribute
        name="mydetector-Temperature"
        type="EPICS_PV"
        source="LINKAM:TEMP"
        dbrtype="DBR_FLOAT"/>
</Attributes>
""",
    )
    with patch("ophyd_async.core._signal.wait_for_value", return_value=None):
        descriptor = await hdf_writer_with_stats.open(DETECTOR_NAME)

    assert descriptor == {
        "test": {
            "source": "mock+ca://HDF:FullFileName_RBV",
            "shape": [1, 10, 10],
            "dtype": "array",
            "dtype_numpy": "<u2",
            "external": "STREAM:",
        },
        "mydetector-sum": {
            "source": "mock+ca://HDF:FullFileName_RBV",
            "shape": [],
            "dtype": "number",
            "dtype_numpy": "<f8",
            "external": "STREAM:",
        },
        "mydetector-Temperature": {
            "dtype": "number",
            "dtype_numpy": "<f4",
            "external": "STREAM:",
            "shape": [],
            "source": "mock+ca://HDF:FullFileName_RBV",
        },
    }
    await hdf_writer_with_stats.close()


async def test_stats_describe_raises_error_with_dbr_native(
    hdf_writer_with_stats: adcore.ADHDFWriter,
):
    assert hdf_writer_with_stats._composer is None
    set_mock_value(hdf_writer_with_stats.fileio.file_path_exists, True)
    assert hdf_writer_with_stats._plugins is not None
    set_mock_value(
        hdf_writer_with_stats._plugins["stats"].nd_attributes_file,
        """<?xml version='1.0' encoding='utf-8'?>
<Attributes>
    <Attribute
        name="mydetector-Temperature"
        type="EPICS_PV"
        source="LINKAM:TEMP"
        dbrtype="DBR_NATIVE"/>
</Attributes>
""",
    )
    with pytest.raises(ValueError) as exc:
        with patch("ophyd_async.core._signal.wait_for_value", return_value=None):
            await hdf_writer_with_stats.open(DETECTOR_NAME)
    await hdf_writer_with_stats.close()
    assert str(exc.value) == "Don't support DBR_NATIVE yet"


async def test_stats_describe_when_plugin_configured_in_memory(RE, detectors):
    for detector in detectors:
        await detector.connect(mock=True)
        detector.set_name(type(detector).__name__)
        RE(setup_ndstats_sum(detector))
        xml = await detector.fileio.nd_attributes_file.get_value()
        root = ET.fromstring(xml)
        for element in root:
            assert str(element.tag) == "Attribute"
            assert element.attrib == {
                "name": f"{detector.name}-sum",
                "type": "PARAM",
                "source": "NDPluginStatsTotal",
                "addr": "0",
                "datatype": "DOUBLE",
                "description": "Sum of the array",
            }


async def test_stats_collect_when_plugin_configured(
    hdf_writer_with_stats: adcore.ADHDFWriter,
):
    set_mock_value(hdf_writer_with_stats.fileio.file_path_exists, True)
    set_mock_value(
        hdf_writer_with_stats._plugins["stats"].nd_attributes_file,
        """<?xml version='1.0' encoding='utf-8'?>
<Attributes>
    <Attribute
        name="mydetector-sum"
        type="PARAM"
        source="TOTAL" addr="0"
        datatype="DOUBLE"
        description="Sum of each detector frame" />
    <Attribute
        name="mydetector-Temperature"
        type="EPICS_PV"
        source="LINKAM:TEMP"
        dbrtype="DBR_FLOAT"/>
</Attributes>
""",
    )
    with patch("ophyd_async.core._signal.wait_for_value", return_value=None):
        await hdf_writer_with_stats.open(DETECTOR_NAME)

    # Check for stream resource documents for the ndattributes
    docs = [
        doc
        async for doc in hdf_writer_with_stats.collect_stream_docs(
            "mydetector", indices_written=3
        )
    ]
    resource_names = [
        doc["data_key"] for name, doc in docs if name == "stream_resource"
    ]
    assert "mydetector-sum" in resource_names
    assert "mydetector-Temperature" in resource_names

    # Check for stream datum documents for the ndattributes
    resource_uids = {doc["uid"] for name, doc in docs if name == "stream_resource"}
    datum_resources = {
        doc["stream_resource"] for name, doc in docs if name == "stream_datum"
    }
    assert resource_uids == datum_resources


async def test_nd_attributes_plan_stub(RE, detectors):
    for detector in detectors:
        await detector.connect(mock=True)
        detector.set_name(type(detector).__name__)
        param = adcore.NDAttributeParam(
            name=f"{detector.name}-sum",
            param="sum",
            datatype=adcore.NDAttributeDataType.DOUBLE,
            description=f"Sum of {detector.name} frame",
        )
        pv = adcore.NDAttributePv(
            name="Temperature",
            signal=epics_signal_r(str, "LINKAM:TEMP"),
            description="The sample temperature",
            dbrtype=adcore.NDAttributePvDbrType.DBR_FLOAT,
        )
        RE(setup_ndattributes(detector.fileio, [pv, param]))
        text = await detector.fileio.nd_attributes_file.get_value()
        xml = ET.fromstring(text)
        assert xml[0].tag == "Attribute"
        assert xml[0].attrib == {
            "name": "Temperature",
            "type": "EPICS_PV",
            "source": "LINKAM:TEMP",
            "dbrtype": "DBR_FLOAT",
            "description": "The sample temperature",
        }

        assert xml[1].tag == "Attribute"
        assert xml[1].attrib == {
            "name": f"{detector.name}-sum",
            "type": "PARAM",
            "source": "sum",
            "addr": "0",
            "datatype": "DOUBLE",
            "description": f"Sum of {detector.name} frame",
        }


async def test_nd_attributes_plan_stub_gives_correct_error(RE, detectors):
    invalid_objects = [0.1, "string", 1, True, False]
    for detector in detectors:
        await detector.connect(mock=True)
        with pytest.raises(ValueError) as exc:
            RE(setup_ndattributes(detector.fileio, invalid_objects))
        assert (
            str(exc.value)
            == f"Invalid type for ndattributes: {type(invalid_objects[0])}. "
            + "Expected NDAttributePv or NDAttributeParam."
        )


async def test_hdf_writer_descriptor_shape_contains_none(
    hdf_writer: adcore.ADHDFWriter,
):
    async def shape_containing_none():
        return 10, None, 10

    hdf_writer._dataset_describer.shape = shape_containing_none

    with pytest.raises(
        ValueError,
        match=r"Datasets with partially unknown dimensionality "
        r"are not currently supported by ADHDFWriter.",
    ):
        await hdf_writer.open(DETECTOR_NAME)



from collections.abc import Mapping, Sequence
from typing import Any, TypeVar, get_args, get_origin

import numpy as np

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    SignalBackend,
    SignalDatatypeT,
    SignalRW,
    StrictEnum,
    SubsetEnum,
    SupersetEnum,
    get_dtype,
    get_enum_cls,
    wait_for_value,
)

T = TypeVar("T")


def get_pv_basename_and_field(pv: str) -> tuple[str, str | None]:
    """Split PV into record name and field."""
    if "." in pv:
        record, field = pv.split(".", maxsplit=1)
    else:
        record, field = pv, None
    return (record, field)


def get_supported_values(
    pv: str,
    datatype: type[T],
    pv_choices: Sequence[str],
) -> Mapping[str, T | str]:
    enum_cls = get_enum_cls(datatype)
    if not enum_cls:
        raise TypeError(f"{datatype} is not an Enum")
    choices = [v.value for v in enum_cls]
    error_msg = f"{pv} has choices {pv_choices}, but {datatype} requested {choices} "
    if issubclass(enum_cls, StrictEnum):
        if set(choices) != set(pv_choices):
            raise TypeError(error_msg + "to be strictly equal to them.")
    elif issubclass(enum_cls, SubsetEnum):
        if not set(choices).issubset(pv_choices):
            raise TypeError(error_msg + "to be a subset of them.")
    elif issubclass(enum_cls, SupersetEnum):
        if not set(pv_choices).issubset(choices):
            raise TypeError(error_msg + "to be a superset of them.")
    else:
        raise TypeError(f"{datatype} is not a StrictEnum, SubsetEnum, or SupersetEnum")
    # Create a map from the string value to the enum instance
    # For StrictEnum and SupersetEnum, all values here will be enum values
    # For SubsetEnum, only the values in choices will be enum values, the rest will be
    # strings
    supported_values = {x: enum_cls(x) for x in pv_choices}
    return supported_values


def format_datatype(datatype: Any) -> str:
    if get_origin(datatype) is np.ndarray and get_args(datatype):
        dtype = get_dtype(datatype)
        return f"Array1D[np.{dtype.name}]"
    elif get_origin(datatype) is Sequence:
        return f"Sequence[{get_args(datatype)[0].__name__}]"
    elif isinstance(datatype, type):
        return datatype.__name__
    else:
        return str(datatype)


class EpicsSignalBackend(SignalBackend[SignalDatatypeT]):
    def __init__(
        self,
        datatype: type[SignalDatatypeT] | None,
        read_pv: str = "",
        write_pv: str = "",
    ):
        self.read_pv = read_pv
        self.write_pv = write_pv
        super().__init__(datatype)


async def stop_busy_record(
    signal: SignalRW[SignalDatatypeT],
    value: SignalDatatypeT,
    timeout: float = DEFAULT_TIMEOUT,
) -> None:
    await signal.set(value, wait=False)
    await wait_for_value(signal, value, timeout=timeout)



from collections.abc import Sequence

import bluesky.plan_stubs as bps

from ophyd_async.epics.adcore import (
    AreaDetector,
    NDArrayBaseIO,
    NDAttributeDataType,
    NDAttributeParam,
    NDAttributePv,
    NDFileHDFIO,
    ndattributes_to_xml,
)


def setup_ndattributes(
    device: NDArrayBaseIO, ndattributes: Sequence[NDAttributeParam | NDAttributePv]
):
    xml = ndattributes_to_xml(ndattributes)
    yield from bps.abs_set(
        device.nd_attributes_file,
        xml,
        wait=True,
    )


def setup_ndstats_sum(detector: AreaDetector):
    """Set up nd stats sum nd attribute for a detector."""
    hdf = getattr(detector, "fileio", None)
    if not isinstance(hdf, NDFileHDFIO):
        msg = (
            f"Expected {detector.name} to have 'fileio' attribute that is an "
            f"NDFileHDFIO, got {hdf}"
        )
        raise TypeError(msg)
    yield from (
        setup_ndattributes(
            hdf,
            [
                NDAttributeParam(
                    name=f"{detector.name}-sum",
                    param="NDPluginStatsTotal",
                    datatype=NDAttributeDataType.DOUBLE,
                    description="Sum of the array",
                )
            ],
        )
    )



"""Test file specifying how we want to interact with the panda controller"""

from unittest.mock import patch

import pytest

from ophyd_async.core import DetectorTrigger, Device, TriggerInfo, init_devices
from ophyd_async.epics.core import epics_signal_rw
from ophyd_async.fastcs.core import fastcs_connector
from ophyd_async.fastcs.panda import CommonPandaBlocks, PandaPcapController


@pytest.fixture
async def mock_panda():
    class Panda(CommonPandaBlocks):
        def __init__(self, uri: str, name: str = ""):
            super().__init__(name=name, connector=fastcs_connector(self, uri))

    async with init_devices(mock=True):
        mock_panda = Panda("PANDACONTROLLER:", name="mock_panda")
        mock_panda.phase_1_signal_units = epics_signal_rw(int, "")
    yield mock_panda


async def test_panda_controller_not_filled_blocks():
    class PcapBlock(Device):
        pass  # Not filled

    controller = PandaPcapController(pcap=PcapBlock())
    with patch("ophyd_async.fastcs.panda._control.wait_for_value", return_value=None):
        with pytest.raises(AttributeError) as exc:
            await controller.prepare(
                TriggerInfo(number_of_events=1, trigger=DetectorTrigger.CONSTANT_GATE)
            )
            await controller.arm()
    assert ("'PcapBlock' object has no attribute 'arm'") in str(exc.value)


async def test_panda_controller_arm_disarm(mock_panda):
    controller = PandaPcapController(mock_panda.pcap)
    with patch("ophyd_async.fastcs.panda._control.wait_for_value", return_value=None):
        await controller.prepare(
            TriggerInfo(number_of_events=1, trigger=DetectorTrigger.CONSTANT_GATE)
        )
        await controller.arm()
        await controller.wait_for_idle()
    await controller.disarm()


async def test_panda_controller_wrong_trigger():
    controller = PandaPcapController(None)
    with pytest.raises(TypeError):
        await controller.prepare(
            TriggerInfo(number_of_events=1, trigger=DetectorTrigger.INTERNAL)
        )



import pytest

from ophyd_async.core import (
    DetectorTrigger,
    TriggerInfo,
)
from ophyd_async.epics import adandor


@pytest.fixture
def test_adandor(ad_standard_det_factory) -> adandor.Andor2Detector:
    return ad_standard_det_factory(adandor.Andor2Detector)


@pytest.mark.parametrize("exposure_time", [0.0, 0.1, 1.0, 10.0, 100.0])
async def test_deadtime_from_exposure_time(
    exposure_time: float,
    test_adandor: adandor.Andor2Detector,
):
    assert test_adandor._controller.get_deadtime(exposure_time) == exposure_time + 0.1


async def test_unsupported_trigger_excepts(test_adandor: adandor.Andor2Detector):
    with pytest.raises(
        ValueError,
        # str(EnumClass.value) handling changed in Python 3.11
        match=("Andor2Controller only supports the following trigger types: .* but"),
    ):
        await test_adandor.prepare(
            TriggerInfo(
                number_of_events=0,
                trigger=DetectorTrigger.VARIABLE_GATE,
                deadtime=1.1,
                livetime=1,
                exposure_timeout=3,
            )
        )



import asyncio

from ophyd_async.core import (
    DetectorTrigger,
    TriggerInfo,
)
from ophyd_async.epics import adcore

from ._kinetix_io import KinetixDriverIO, KinetixTriggerMode

KINETIX_TRIGGER_MODE_MAP = {
    DetectorTrigger.INTERNAL: KinetixTriggerMode.INTERNAL,
    DetectorTrigger.CONSTANT_GATE: KinetixTriggerMode.GATE,
    DetectorTrigger.VARIABLE_GATE: KinetixTriggerMode.GATE,
    DetectorTrigger.EDGE_TRIGGER: KinetixTriggerMode.EDGE,
}


class KinetixController(adcore.ADBaseController[KinetixDriverIO]):
    """Controller for adkinetix detector."""

    def __init__(
        self,
        driver: KinetixDriverIO,
        good_states: frozenset[adcore.ADState] = adcore.DEFAULT_GOOD_STATES,
    ) -> None:
        super().__init__(driver, good_states=good_states)

    def get_deadtime(self, exposure: float | None) -> float:
        return 0.001

    async def prepare(self, trigger_info: TriggerInfo):
        await asyncio.gather(
            self.driver.trigger_mode.set(
                KINETIX_TRIGGER_MODE_MAP[trigger_info.trigger]
            ),
            self.driver.num_images.set(trigger_info.total_number_of_exposures),
            self.driver.image_mode.set(adcore.ADImageMode.MULTIPLE),
        )
        if trigger_info.livetime is not None and trigger_info.trigger not in [
            DetectorTrigger.VARIABLE_GATE,
            DetectorTrigger.CONSTANT_GATE,
        ]:
            await self.driver.acquire_time.set(trigger_info.livetime)



"""Top level API."""

from . import core
from ._version import version

__version__ = version
"""Version number as calculated by https://github.com/pypa/setuptools_scm"""

__all__ = ["__version__", "core"]



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Iterator, MutableMapping
from typing import Any, Generic

from ._device import Device, DeviceT
from ._signal import SignalRW
from ._signal_backend import SignalDatatypeT


class Settings(MutableMapping[SignalRW[Any], Any], Generic[DeviceT]):
    """Used for supplying settings to signals.

    :param device: The device that the settings are for.
    :param settings: A dictionary of settings to start with.

    :example:
    ```python
    # Settings are created from a dict of signals to values
    settings1 = Settings(device, {device.sig1: 1, device.sig2: 2})
    settings2 = Settings(device, {device.sig1: 10, device.sig3: 3})
    # They act like a dictionaries
    assert settings1[device.sig1] == 1
    # Including the ability to "or" two settings together
    settings = settings1 | settings2
    assert dict(settings) == {
        device.sig1: 10,
        device.sig2: 2,
        device.sig3: 3,
    }
    ```
    """

    def __init__(
        self, device: DeviceT, settings: MutableMapping[SignalRW, Any] | None = None
    ):
        self.device = device
        self._settings = {}
        self.update(settings or {})

    def __getitem__(self, key: SignalRW[SignalDatatypeT]) -> SignalDatatypeT:
        return self._settings[key]

    def _is_in_device(self, device: Device) -> bool:
        while device.parent and device.parent is not self.device:
            # While we have a parent that is not the right device
            # continue searching up the tree
            device = device.parent
        return device.parent is self.device

    def __setitem__(
        self, key: SignalRW[SignalDatatypeT], value: SignalDatatypeT | None
    ) -> None:
        # Check the types on entry to dict to make sure we can't accidentally
        # add a non-signal type
        if not isinstance(key, SignalRW):
            raise TypeError(f"Expected SignalRW, got {key}")
        if not self._is_in_device(key):
            raise KeyError(f"Signal {key} is not a child of {self.device}")
        self._settings[key] = value

    def __delitem__(self, key: SignalRW) -> None:
        del self._settings[key]

    def __iter__(self) -> Iterator[SignalRW]:
        yield from iter(self._settings)

    def __len__(self) -> int:
        return len(self._settings)

    def __or__(self, other: MutableMapping[SignalRW, Any]) -> Settings[DeviceT]:
        """Create a new Settings that is the union of self overridden by other."""
        if isinstance(other, Settings) and not self._is_in_device(other.device):
            raise ValueError(f"{other.device} is not a child of {self.device}")
        return Settings(self.device, self._settings | dict(other))

    def partition(
        self, predicate: Callable[[SignalRW], bool]
    ) -> tuple[Settings[DeviceT], Settings[DeviceT]]:
        """Partition into two Settings based on a predicate.

        :param predicate:
            Callable that takes each signal, and returns a boolean to say if it
            should be in the first returned Settings
        :returns:
            `(where_true, where_false)` where each is a Settings object.
            The first contains the signals for which the predicate returned True,
            and the second contains the signals for which the predicate returned False.

        :example:
        ```python
        settings = Settings(device, {device.special: 1, device.sig: 2})
        specials, others = settings.partition(lambda sig: "special" in sig.name)
        ```
        """
        where_true, where_false = Settings(self.device), Settings(self.device)
        for signal, value in self.items():
            dest = where_true if predicate(signal) else where_false
            dest[signal] = value
        return where_true, where_false


class SettingsProvider:
    """Base class for providing settings."""

    @abstractmethod
    async def store(self, name: str, data: dict[str, Any]):
        """Store the data, associating it with the given name."""

    @abstractmethod
    async def retrieve(self, name: str) -> dict[str, Any]:
        """Retrieve the data associated with the given name."""



from typing import Any, Generic

import numpy as np
from numpy.typing import NDArray
from tango import DevState

from ophyd_async.core import SignalDatatypeT

from ._utils import DevStateEnum


class TangoConverter(Generic[SignalDatatypeT]):
    def write_value(self, value: Any) -> Any:
        return value

    def value(self, value: Any) -> Any:
        return value


class TangoEnumConverter(TangoConverter):
    def __init__(self, labels: list[str]):
        self._labels = labels

    def write_value(self, value: str):
        if not isinstance(value, str):
            raise TypeError("TangoEnumConverter expects str value")
        return self._labels.index(value)

    def value(self, value: int):
        return self._labels[value]


class TangoEnumArrayConverter(TangoConverter):
    def __init__(self, labels: list[str]):
        self._labels = labels

    def write_value(self, value: NDArray[np.str_]) -> NDArray[np.integer]:
        vfunc = np.vectorize(self._labels.index)
        new_array = vfunc(value)
        return new_array

    def value(self, value: NDArray[np.integer]) -> NDArray[np.str_]:
        vfunc = np.vectorize(self._labels.__getitem__)
        new_array = vfunc(value)
        return new_array


class TangoDevStateConverter(TangoConverter):
    _labels = [e.value for e in DevStateEnum]

    def write_value(self, value: str) -> DevState:
        idx = self._labels.index(value)
        return DevState(idx)

    def value(self, value: DevState) -> str:
        idx = int(value)
        return self._labels[idx]


class TangoDevStateArrayConverter(TangoConverter):
    _labels = [e.value for e in DevStateEnum]

    def _write_convert(self, value):
        return DevState(self._labels.index(value))

    def _convert(self, value):
        return self._labels[int(value)]

    def write_value(self, value: NDArray[np.str_]) -> NDArray[DevState]:
        vfunc = np.vectorize(self._write_convert, otypes=[DevState])
        new_array = vfunc(value)
        return new_array

    def value(self, value: NDArray[DevState]) -> NDArray[np.str_]:
        vfunc = np.vectorize(self._convert)
        new_array = vfunc(value)
        return new_array



import asyncio
from dataclasses import dataclass

import numpy as np
from scanspec.core import Path, Slice
from scanspec.specs import Spec

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    FlyerController,
    error_if_none,
    observe_value,
    set_and_wait_for_value,
    wait_for_value,
)
from ophyd_async.epics.motor import Motor
from ophyd_async.epics.pmac import PmacIO
from ophyd_async.epics.pmac._pmac_io import CS_LETTERS
from ophyd_async.epics.pmac._pmac_trajectory_generation import PVT, Trajectory
from ophyd_async.epics.pmac._utils import (
    _PmacMotorInfo,
    calculate_ramp_position_and_duration,
)

# PMAC durations are in milliseconds
# We must convert from scanspec durations (seconds) to milliseconds
# PMAC motion program multiples durations by 0.001
# (see https://github.com/DiamondLightSource/pmac/blob/afe81f8bb9179c3a20eff351f30bc6cfd1539ad9/pmacApp/pmc/trajectory_scan_code_ppmac.pmc#L241)
# Therefore, we must divide scanspec durations by 10e-6
TICK_S = 0.000001
SLICE_SIZE = 4000


@dataclass
class PmacPrepareContext:
    path: Path
    motor_info: _PmacMotorInfo
    ramp_up_time: float


class PmacTrajectoryTriggerLogic(FlyerController):
    def __init__(self, pmac: PmacIO) -> None:
        self.pmac = pmac
        self._next_pvt: PVT | None
        self._loaded: int = 0
        self._trajectory_status: AsyncStatus | None = None
        self._prepare_context: PmacPrepareContext | None = None

    async def prepare(self, value: Spec[Motor]):
        path = Path(value.calculate())
        slice = path.consume(SLICE_SIZE)
        path_length = len(path)
        motors = slice.axes()
        motor_info = await _PmacMotorInfo.from_motors(self.pmac, motors)
        ramp_up_pos, ramp_up_time = calculate_ramp_position_and_duration(
            slice, motor_info, True
        )
        self._prepare_context = PmacPrepareContext(
            path=path, motor_info=motor_info, ramp_up_time=ramp_up_time
        )
        await asyncio.gather(
            self._build_trajectory(motor_info, slice, path_length, ramp_up_time),
            self._move_to_start(motor_info, ramp_up_pos),
        )

    async def kickoff(self):
        prepare_context = error_if_none(
            self._prepare_context, "Cannot kickoff. Must call prepare first."
        )
        self._prepare_context = None
        self._trajectory_status = self._execute_trajectory(
            prepare_context.path, prepare_context.motor_info
        )
        # Wait for the ramp up to happen
        await wait_for_value(
            self.pmac.trajectory.total_points,
            lambda v: v >= 1,
            prepare_context.ramp_up_time + DEFAULT_TIMEOUT,
        )

    async def complete(self):
        trajectory_status = error_if_none(
            self._trajectory_status, "Cannot complete. Must call kickoff first."
        )
        await trajectory_status
        # Reset trajectory status and number of loaded points
        self._trajectory_status = None
        self._loaded = 0

    async def stop(self):
        await self.pmac.trajectory.abort_profile.trigger()

    @AsyncStatus.wrap
    async def _execute_trajectory(self, path: Path, motor_info: _PmacMotorInfo):
        execute_status = self.pmac.trajectory.execute_profile.set(
            True, wait=True, timeout=None
        )
        # We consume SLICE_SIZE from self.path and parse a trajectory
        # containing at least 2 * SLICE_SIZE, as a gapless trajectory
        # will contain 2 points per slice frame. If gaps are present,
        # additional points are inserted, overfilling the buffer.
        min_buffer_size = SLICE_SIZE * 2
        async for current_point in observe_value(
            self.pmac.trajectory.total_points,
            done_status=execute_status,
            timeout=DEFAULT_TIMEOUT,
        ):
            # Ensure we maintain a minimum buffer size, if we have more points to append
            if len(path) != 0 and self._loaded - current_point < min_buffer_size:
                # We have less than SLICE_SIZE * 2 points in the buffer, so refill
                next_slice = path.consume(SLICE_SIZE)
                path_length = len(path)
                await self._append_trajectory(next_slice, path_length, motor_info)

    async def _append_trajectory(
        self, slice: Slice, path_length: int, motor_info: _PmacMotorInfo
    ):
        trajectory = await self._parse_trajectory(slice, path_length, motor_info)
        await self._set_trajectory_arrays(trajectory, motor_info)
        await self.pmac.trajectory.append_profile.trigger()

    async def _build_trajectory(
        self,
        motor_info: _PmacMotorInfo,
        slice: Slice,
        path_length: int,
        ramp_up_time: float,
    ):
        trajectory = await self._parse_trajectory(
            slice, path_length, motor_info, ramp_up_time
        )
        use_axis = {
            axis + 1: (axis in motor_info.motor_cs_index.values())
            for axis in range(len(CS_LETTERS))
        }

        coros = [
            self.pmac.trajectory.profile_cs_name.set(motor_info.cs_port),
            self.pmac.trajectory.calculate_velocities.set(False),
            self._set_trajectory_arrays(trajectory, motor_info),
        ] + [
            self.pmac.trajectory.use_axis[number].set(use)
            for number, use in use_axis.items()
        ]

        await asyncio.gather(*coros)
        await self.pmac.trajectory.build_profile.trigger()

    async def _parse_trajectory(
        self,
        slice: Slice,
        path_length: int,
        motor_info: _PmacMotorInfo,
        ramp_up_time: float | None = None,
    ) -> Trajectory:
        trajectory, exit_pvt = Trajectory.from_slice(
            slice,
            motor_info,
            None if ramp_up_time else self._next_pvt,
            ramp_up_time=ramp_up_time,
        )

        if path_length == 0:
            ramp_down_pos, ramp_down_time = calculate_ramp_position_and_duration(
                slice, motor_info, False
            )
            trajectory = trajectory.with_ramp_down(
                exit_pvt, ramp_down_pos, ramp_down_time, 0
            )
        self._next_pvt = exit_pvt
        self._loaded += len(trajectory)

        return trajectory

    async def _set_trajectory_arrays(
        self, trajectory: Trajectory, motor_info: _PmacMotorInfo
    ):
        coros = []
        for motor, number in motor_info.motor_cs_index.items():
            coros.append(
                self.pmac.trajectory.positions[number + 1].set(
                    trajectory.positions[motor]
                )
            )
            coros.append(
                self.pmac.trajectory.velocities[number + 1].set(
                    trajectory.velocities[motor]
                )
            )
        coros.extend(
            [
                self.pmac.trajectory.time_array.set(trajectory.durations / TICK_S),
                self.pmac.trajectory.user_array.set(trajectory.user_programs),
                self.pmac.trajectory.points_to_build.set(len(trajectory.durations)),
            ]
        )
        await asyncio.gather(*coros)

    async def _move_to_start(
        self, motor_info: _PmacMotorInfo, ramp_up_position: dict[Motor, np.float64]
    ):
        coord = self.pmac.coord[motor_info.cs_number]
        coros = []
        await coord.defer_moves.set(True)
        for motor, position in ramp_up_position.items():
            coros.append(
                set_and_wait_for_value(
                    coord.cs_axis_setpoint[motor_info.motor_cs_index[motor] + 1],
                    position,
                    set_timeout=10,
                    wait_for_set_completion=False,
                )
            )
        statuses = await asyncio.gather(*coros)
        await coord.defer_moves.set(False)
        await asyncio.gather(*statuses)



from __future__ import annotations

import asyncio
import time
from pathlib import PurePath

import h5py
import numpy as np

# raw data path
DATA_PATH = "/entry/data/data"

# pixel sum path
SUM_PATH = "/entry/sum"


def generate_gaussian_blob(height: int, width: int) -> np.ndarray:
    """Make a Gaussian Blob with float values in range 0..1."""
    x, y = np.meshgrid(np.linspace(-1, 1, width), np.linspace(-1, 1, height))
    d = np.sqrt(x * x + y * y)
    blob = np.exp(-(d**2))
    return blob


def generate_interesting_pattern(
    x: float, y: float, channel: int, offset: float
) -> float:
    """Return a float value in range 0..1.

    Interesting in x and y in range -10..10
    """
    return (np.sin(x) ** channel + np.cos(x * y + offset) + 2) / 4


class PatternFile:
    def __init__(
        self,
        path: PurePath,
        width: int = 320,
        height: int = 240,
    ):
        self.file = h5py.File(path, "w", libver="latest")
        self.data = self.file.create_dataset(
            name=DATA_PATH,
            shape=(0, height, width),
            dtype=np.uint8,
            maxshape=(None, height, width),
            chunks=(1024, height, width),
        )
        self.sum = self.file.create_dataset(
            name=SUM_PATH,
            shape=(0,),
            dtype=np.int64,
            maxshape=(None,),
            chunks=(1024,),
        )
        # Once datasets written, can switch the model to single writer multiple reader
        self.file.swmr_mode = True
        self.blob = generate_gaussian_blob(height, width) * np.iinfo(np.uint8).max
        self.image_counter = 0
        self.e = asyncio.Event()

    def write_image_to_file(self, intensity: float):
        data = np.floor(self.blob * intensity)
        for dset, value in ((self.data, data), (self.sum, np.sum(data))):
            dset.resize(self.image_counter + 1, axis=0)
            dset[self.image_counter] = value
            dset.flush()
        self.image_counter += 1
        self.e.set()
        self.e.clear()

    def close(self):
        self.file.close()


class PatternGenerator:
    """Generates pattern images in files."""

    def __init__(self, sleep=asyncio.sleep):
        self._x = 0.0
        self._y = 0.0
        self._file: PatternFile | None = None
        self.sleep = sleep

    def set_x(self, x: float):
        self._x = x

    def set_y(self, y: float):
        self._y = y

    def generate_point(self, channel: int = 1, high_energy: bool = False) -> float:
        """Make a point between 0 and 1 based on x and y."""
        offset = 100 if high_energy else 10
        return generate_interesting_pattern(self._x, self._y, channel, offset)

    def open_file(self, path: PurePath, width: int, height: int):
        self._file = PatternFile(path, width, height)

    def _get_file(self) -> PatternFile:
        if not self._file:
            raise RuntimeError("open_file not run")
        return self._file

    async def write_images_to_file(
        self, exposure: float, period: float, number_of_frames: int
    ):
        file = self._get_file()
        start = time.monotonic()
        for i in range(1, number_of_frames + 1):
            deadline = start + i * period
            timeout = deadline - time.monotonic()
            await self.sleep(timeout)
            intensity = self.generate_point() * exposure
            file.write_image_to_file(intensity)

    async def wait_for_next_index(self, timeout: float):
        await asyncio.wait_for(self._get_file().e.wait(), timeout)

    def get_last_index(self) -> int:
        return self._get_file().image_counter

    def close_file(self):
        if self._file:
            self._file.close()
            self._file = None



import asyncio
from collections.abc import AsyncGenerator, AsyncIterator

from bluesky.protocols import StreamAsset
from event_model import DataKey

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    DetectorWriter,
    HDFDatasetDescription,
    HDFDocumentComposer,
    PathProvider,
    observe_value,
    wait_for_value,
)

from ._block import DataBlock, PandaCaptureMode


class PandaHDFWriter(DetectorWriter):
    """For writing for PandA data from the `DataBlock`."""

    def __init__(
        self,
        path_provider: PathProvider,
        panda_data_block: DataBlock,
    ) -> None:
        self.panda_data_block = panda_data_block
        self._path_provider = path_provider
        self._datasets: list[HDFDatasetDescription] = []
        self._composer: HDFDocumentComposer | None = None

    # Triggered on PCAP arm
    async def open(self, name: str, exposures_per_event: int = 1) -> dict[str, DataKey]:
        self._composer = None
        """Retrieve and get descriptor of all PandA signals marked for capture."""
        self._exposures_per_event = exposures_per_event
        # Ensure flushes are immediate
        await self.panda_data_block.flush_period.set(0)

        info = self._path_provider(device_name=name)

        # Set create dir depth first to guarantee that callback when setting
        # directory path has correct value
        await self.panda_data_block.create_directory.set(info.create_dir_depth)

        # Set the initial values
        await asyncio.gather(
            self.panda_data_block.hdf_directory.set(str(info.directory_path)),
            self.panda_data_block.hdf_file_name.set(
                f"{info.filename}.h5",
            ),
            self.panda_data_block.capture_mode.set(PandaCaptureMode.FOREVER),
        )

        # Make sure that directory exists or has been created.
        if not await self.panda_data_block.directory_exists.get_value() == 1:
            raise OSError(
                f"Directory {info.directory_path} does not exist or "
                "is not writable by the PandABlocks-ioc!"
            )

        # Wait for it to start, stashing the status that tells us when it finishes
        await self.panda_data_block.capture.set(True)

        describe = await self._describe(name)

        self._composer = HDFDocumentComposer(
            f"{info.directory_uri}{info.filename}.h5",
            self._datasets,
        )

        return describe

    async def _describe(self, name: str) -> dict[str, DataKey]:
        """Return a describe based on the datasets PV."""
        await self._update_datasets(name)
        describe = {
            ds.data_key: DataKey(
                source=self.panda_data_block.hdf_directory.source,
                shape=list(ds.shape),
                dtype="array"
                if self._exposures_per_event > 1 or len(ds.shape) > 1
                else "number",
                # PandA data should always be written as Float64
                dtype_numpy=ds.dtype_numpy,
                external="STREAM:",
            )
            for ds in self._datasets
        }
        return describe

    async def _update_datasets(self, name: str) -> None:
        # Load data from the datasets PV on the panda, update internal
        # representation of datasets that the panda will write.
        capture_table = await self.panda_data_block.datasets.get_value()
        self._datasets = [
            # TODO: Update chunk size to read signal once available in IOC
            # Currently PandA IOC sets chunk size to 1024 points per chunk
            HDFDatasetDescription(
                data_key=dataset_name,
                dataset="/" + dataset_name,
                shape=(self._exposures_per_event,)
                if self._exposures_per_event > 1
                else (),
                dtype_numpy="<f8",
                chunk_shape=(1024,),
            )
            for dataset_name in capture_table.name
        ]

        # Warn user if dataset table is empty in PandA
        # i.e. no stream resources will be generated
        if len(self._datasets) == 0:
            self.panda_data_block.log.warning(
                f"PandA {name} DATASETS table is empty! "
                "No stream resource docs will be generated. "
                "Make sure captured positions have their corresponding "
                "*:DATASET PV set to a scientifically relevant name."
            )

    # Next few functions are exactly the same as AD writer. Could move as default
    # StandardDetector behavior
    async def wait_for_index(self, index: int, timeout: float | None = DEFAULT_TIMEOUT):
        def matcher(value: int) -> bool:
            # Index is already divided by exposures_per_event, so we need to also
            # divide the value by exposures_per_event to get the correct index
            return value // self._exposures_per_event >= index

        matcher.__name__ = f"index_at_least_{index}"
        await wait_for_value(
            self.panda_data_block.num_captured, matcher, timeout=timeout
        )

    async def get_indices_written(self) -> int:
        return (
            await self.panda_data_block.num_captured.get_value()
            // self._exposures_per_event
        )

    async def observe_indices_written(
        self, timeout: float
    ) -> AsyncGenerator[int, None]:
        """Wait until a specific index is ready to be collected."""
        async for num_captured in observe_value(
            self.panda_data_block.num_captured, timeout
        ):
            yield num_captured // self._exposures_per_event

    async def collect_stream_docs(
        self, name: str, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        # TODO: fail if we get dropped frames
        if self._composer is None:
            msg = f"open() not called on {self}"
            raise RuntimeError(msg)
        for doc in self._composer.make_stream_docs(indices_written):
            yield doc

    # Could put this function as default for StandardDetector
    async def close(self):
        await self.panda_data_block.capture.set(
            False, wait=True, timeout=DEFAULT_TIMEOUT
        )



from ophyd_async.core import (
    Device,
    SignalR,
    SignalRW,
    SignalX,
    StrictEnum,
)
from ophyd_async.fastcs.core import fastcs_connector


class EigerTriggerMode(StrictEnum):
    INTERNAL = "ints"
    EDGE = "exts"
    GATE = "exte"


class EigerMonitorIO(Device):
    pass


class EigerStreamIO(Device):
    pass


class EigerDetectorIO(Device):
    bit_depth_image: SignalR[int]
    state: SignalR[str]
    count_time: SignalRW[float]
    frame_time: SignalRW[float]
    nimages: SignalRW[int]
    ntrigger: SignalRW[int]
    nexpi: SignalRW[int]
    trigger_mode: SignalRW[str]
    roi_mode: SignalRW[str]
    photon_energy: SignalRW[float]
    beam_center_x: SignalRW[float]
    beam_center_y: SignalRW[float]
    detector_distance: SignalRW[float]
    omega_start: SignalRW[float]
    omega_increment: SignalRW[float]
    arm: SignalX
    disarm: SignalX
    trigger: SignalX


class EigerDriverIO(Device):
    """Contains signals for handling IO on the Eiger detector."""

    stale_parameters: SignalR[bool]
    monitor: EigerMonitorIO
    stream: EigerStreamIO
    detector: EigerDetectorIO

    def __init__(self, uri: str, name: str = ""):
        super().__init__(name=name, connector=fastcs_connector(self, uri))



from abc import abstractmethod
from collections.abc import Sequence
from typing import Generic, TypedDict, TypeVar, get_origin

import numpy as np
from bluesky.protocols import Reading
from event_model import DataKey, Dtype, Limits

from ophyd_async.core._utils import (
    Callback,
    EnumTypes,
    StrictEnum,
    SubsetEnum,
    SupersetEnum,
    get_enum_cls,
)

from ._table import Table

DTypeScalar_co = TypeVar("DTypeScalar_co", covariant=True, bound=np.generic)
"""A numpy dtype like [](#numpy.float64)."""


# To be a 1D array shape should really be tuple[int], but np.array()
# currently produces tuple[int, ...] even when it has 1D input args
# https://github.com/numpy/numpy/issues/28077#issuecomment-2566485178
Array1D = np.ndarray[tuple[int, ...], np.dtype[DTypeScalar_co]]
"""A type alias for a 1D numpy array with a specific scalar data type.

E.g. `Array1D[np.float64]` is a 1D numpy array of 64-bit floats."""

Primitive = bool | int | float | str
SignalDatatype = (
    Primitive
    | EnumTypes
    | Array1D[np.bool_]
    | Array1D[np.int8]
    | Array1D[np.uint8]
    | Array1D[np.int16]
    | Array1D[np.uint16]
    | Array1D[np.int32]
    | Array1D[np.uint32]
    | Array1D[np.int64]
    | Array1D[np.uint64]
    | Array1D[np.float32]
    | Array1D[np.float64]
    | np.ndarray
    | Sequence[str]
    | Sequence[StrictEnum]
    | Sequence[SubsetEnum]
    | Sequence[SupersetEnum]
    | Table
)
"""The supported [](#Signal) datatypes:

- A python primitive [](#bool), [](#int), [](#float), [](#str)
- An [](#EnumTypes) subclass
- A fixed datatype [](#Array1D) of numpy bool, signed and unsigned integers or float
- A [](#numpy.ndarray) which can change dimensions and datatype at runtime
- A sequence of [](#str)
- A sequence of [](#EnumTypes) subclasses
- A [](#Table) subclass
"""
# TODO: These typevars will not be needed when we drop python 3.11
# as you can do MyConverter[SignalType: SignalTypeUnion]:
# rather than MyConverter(Generic[SignalType])
PrimitiveT = TypeVar("PrimitiveT", bound=Primitive)
SignalDatatypeT = TypeVar("SignalDatatypeT", bound=SignalDatatype)
"""A typevar for a [](#SignalDatatype)."""
SignalDatatypeV = TypeVar("SignalDatatypeV", bound=SignalDatatype)
EnumT = TypeVar("EnumT", bound=EnumTypes)
TableT = TypeVar("TableT", bound=Table)


class SignalBackend(Generic[SignalDatatypeT]):
    """A read/write/monitor backend for a Signals."""

    def __init__(self, datatype: type[SignalDatatypeT] | None):
        self.datatype = datatype

    @abstractmethod
    def source(self, name: str, read: bool) -> str:
        """Return source of signal.

        :param name: The name of the signal, which can be used or discarded.
        :param read: True if we want the source for reading, False if writing.
        """

    @abstractmethod
    async def connect(self, timeout: float):
        """Connect to underlying hardware."""

    @abstractmethod
    async def put(self, value: SignalDatatypeT | None, wait: bool):
        """Put a value to the PV, if wait then wait for completion."""

    @abstractmethod
    async def get_datakey(self, source: str) -> DataKey:
        """Metadata like source, dtype, shape, precision, units."""

    @abstractmethod
    async def get_reading(self) -> Reading[SignalDatatypeT]:
        """Return the current value, timestamp and severity."""

    @abstractmethod
    async def get_value(self) -> SignalDatatypeT:
        """Return the current value."""

    @abstractmethod
    async def get_setpoint(self) -> SignalDatatypeT:
        """Return the point that a signal was requested to move to."""

    @abstractmethod
    def set_callback(self, callback: Callback[Reading[SignalDatatypeT]] | None) -> None:
        """Observe changes to the current value, timestamp and severity."""


_primitive_dtype: dict[type[Primitive], Dtype] = {
    bool: "boolean",
    int: "integer",
    float: "number",
    str: "string",
}


class SignalMetadata(TypedDict, total=False):
    """Metadata for a signal. No field is required."""

    limits: Limits
    """The control, display, warning and alarm limits for a numeric datatype."""

    choices: list[str]
    """The choice of possible values for an enum datatype."""

    precision: int
    """The number of digits after the decimal place to display for a float datatype."""

    units: str
    """The engineering units of the value for a numeric datatype."""


def _datakey_dtype(datatype: type[SignalDatatype]) -> Dtype:
    if (
        datatype is np.ndarray
        or get_origin(datatype) in (Sequence, np.ndarray)
        or issubclass(datatype, Table)
    ):
        return "array"
    elif issubclass(datatype, EnumTypes):
        return "string"
    elif issubclass(datatype, Primitive):
        return _primitive_dtype[datatype]
    else:
        raise TypeError(f"Can't make dtype for {datatype}")


def _datakey_dtype_numpy(
    datatype: type[SignalDatatypeT], value: SignalDatatypeT
) -> np.dtype:
    if isinstance(value, np.ndarray):
        # The value already has a dtype, use that
        return value.dtype
    elif (
        get_origin(datatype) is Sequence
        or datatype is str
        or issubclass(datatype, EnumTypes)
    ):
        # TODO: use np.dtypes.StringDType when we can use in structured arrays
        # https://github.com/numpy/numpy/issues/25693
        return np.dtype("S40")
    elif isinstance(value, Table):
        return value.numpy_dtype()
    elif issubclass(datatype, Primitive):
        return np.dtype(datatype)
    else:
        raise TypeError(f"Can't make dtype_numpy for {datatype}")


def _datakey_shape(value: SignalDatatype) -> list[int | None]:
    if type(value) in _primitive_dtype or isinstance(value, EnumTypes):
        return []
    elif isinstance(value, np.ndarray):
        return list(value.shape)
    elif isinstance(value, Sequence | Table):
        return [len(value)]
    else:
        raise TypeError(
            f"Can't make shape for {value} with SignalDataType: {type(value)}"
        )


def make_datakey(
    datatype: type[SignalDatatypeT],
    value: SignalDatatypeT,
    source: str,
    metadata: SignalMetadata,
) -> DataKey:
    """Make a DataKey for a given datatype."""
    dtn = _datakey_dtype_numpy(datatype, value)
    return DataKey(
        dtype=_datakey_dtype(datatype),
        shape=_datakey_shape(value),
        # Ignore until https://github.com/bluesky/event-model/issues/308
        dtype_numpy=dtn.descr if len(dtn.descr) > 1 else dtn.str,  # type: ignore
        source=source,
        **metadata,
    )


def make_metadata(
    datatype: type[SignalDatatypeT] | None,
    units: str | None = None,
    precision: int | None = None,
) -> SignalMetadata:
    metadata: SignalMetadata = {}
    if units is not None:
        metadata["units"] = units
    if precision is not None:
        metadata["precision"] = precision
    if enum_cls := get_enum_cls(datatype):
        metadata["choices"] = [v.value for v in enum_cls]
    return metadata



import asyncio
import time
import traceback
from unittest.mock import MagicMock, Mock

import pytest

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    Device,
    DeviceVector,
    NotConnectedError,
    Reference,
    SignalRW,
    init_devices,
    soft_signal_rw,
    wait_for_connection,
)
from ophyd_async.epics import motor
from ophyd_async.plan_stubs import ensure_connected


class DummyBaseDevice(Device):
    def __init__(self) -> None:
        self.connected = False
        super().__init__()

    async def connect(
        self, mock=False, timeout=DEFAULT_TIMEOUT, force_reconnect: bool = False
    ):
        self.connected = True


class DummyDeviceGroup(Device):
    def __init__(self, name: str) -> None:
        self.child1 = DummyBaseDevice()
        self._child2 = DummyBaseDevice()
        self.dict_with_children: DeviceVector[DummyBaseDevice] = DeviceVector(
            {123: DummyBaseDevice()}
        )
        super().__init__(name)


@pytest.fixture
def parent() -> DummyDeviceGroup:
    return DummyDeviceGroup("parent")


class DeviceWithNamedChild(Device):
    def __init__(self, name: str = "") -> None:
        super().__init__(name)
        self.child = soft_signal_rw(int, name="foo")


def test_device_signal_naming():
    device = DeviceWithNamedChild("bar")
    assert device.name == "bar"
    assert device.child.name == "bar-child"


class DeviceWithRefToSignal(Device):
    def __init__(self, signal: SignalRW[int]):
        self.signal_ref = Reference(signal)
        super().__init__(name="bat")

    def get_source(self) -> str:
        return self.signal_ref().source


async def test_device_connect_missing_connector() -> None:
    # Create an instance of Device without calling __init__
    device = object.__new__(Device)
    # assert the init method wasn't called
    assert hasattr(device, "_connector") is False
    with pytest.raises(
        RuntimeError,
        match=r".* doesn't have attribute `_connector`.*",
    ):
        await device.connect(mock=True)


def test_device_with_signal_ref_does_not_rename():
    device = DeviceWithNamedChild()
    device.set_name("bar")
    assert dict(device.children()) == {"child": device.child}
    private_device = DeviceWithRefToSignal(device.child)
    assert device.child.source == private_device.get_source()
    assert dict(private_device.children()) == {}
    assert device.name == "bar"
    assert device.child.name == "bar-child"
    assert private_device.name == "bat"


def test_device_children(parent: DummyDeviceGroup):
    names = ["child1", "_child2", "dict_with_children"]
    for idx, (name, child) in enumerate(parent.children()):
        assert name == names[idx]
        expected_type = (
            DeviceVector if name == "dict_with_children" else DummyBaseDevice
        )
        assert type(child) is expected_type
        assert child.parent == parent


def test_device_vector_children():
    parent = DummyDeviceGroup("root")

    device_vector_children = list(parent.dict_with_children.items())
    assert device_vector_children == [(123, parent.dict_with_children[123])]


async def test_children_of_device_have_set_names_and_get_connected(
    parent: DummyDeviceGroup,
):
    assert parent.name == "parent"
    assert parent.child1.name == "parent-child1"
    assert parent._child2.name == "parent-_child2"
    assert parent.dict_with_children.name == "parent-dict_with_children"
    assert parent.dict_with_children[123].name == "parent-dict_with_children-123"

    await parent.connect()

    assert parent.child1.connected
    assert parent.dict_with_children[123].connected


async def test_children_of_device_with_different_separator(
    parent: DummyDeviceGroup,
):
    for separator in ("_", None):
        # The second time round, check that it doesn't change name if
        # we pass None, as this is what PviConnector does
        parent.set_name("parent", child_name_separator=separator)
        assert parent.name == "parent"
        assert parent.child1.name == "parent_child1"
        assert parent._child2.name == "parent__child2"
        assert parent.dict_with_children.name == "parent_dict_with_children"
        assert parent.dict_with_children[123].name == "parent_dict_with_children_123"


async def test_device_with_init_devices():
    async with init_devices(mock=True):
        parent = DummyDeviceGroup("parent")

    assert parent.name == "parent"
    assert parent.parent is None
    assert parent.child1.name == "parent-child1"
    assert parent.child1.parent == parent
    assert parent._child2.name == "parent-_child2"
    assert parent._child2.parent == parent
    assert parent.dict_with_children.name == "parent-dict_with_children"
    assert parent.dict_with_children.parent == parent
    assert parent.dict_with_children[123].name == "parent-dict_with_children-123"
    assert parent.dict_with_children[123].parent == parent.dict_with_children
    assert parent.child1.connected
    assert parent.dict_with_children[123].connected


async def test_wait_for_connection():
    class DummyDeviceWithSleep(DummyBaseDevice):
        def __init__(self, name) -> None:
            self.set_name(name)

        async def connect(self, mock=False, timeout=DEFAULT_TIMEOUT):
            await asyncio.sleep(0.01)
            self.connected = True

    device1, device2 = DummyDeviceWithSleep("device1"), DummyDeviceWithSleep("device2")

    normal_coros = {"device1": device1.connect(), "device2": device2.connect()}

    await wait_for_connection(**normal_coros)

    assert device1.connected
    assert device2.connected


async def test_wait_for_connection_propagates_error(
    normal_coroutine, failing_coroutine
):
    failing_coros = {"test": normal_coroutine(), "failing": failing_coroutine()}

    with pytest.raises(NotConnectedError) as exc:
        await wait_for_connection(**failing_coros)
        assert traceback.extract_tb(exc.__traceback__)[-1].name == "failing_coroutine"


async def test_device_log_has_correct_name():
    device = DummyBaseDevice()
    assert device.log.extra["ophyd_async_device_name"] == ""
    device.set_name("device")
    assert device.log.extra["ophyd_async_device_name"] == "device"


class MotorBundle(Device):
    def __init__(self, name: str) -> None:
        self.X = motor.Motor("BLxxI-MO-TABLE-01:X")
        self.Y = motor.Motor("BLxxI-MO-TABLE-01:Y")
        self.V: DeviceVector[motor.Motor] = DeviceVector(
            {
                0: motor.Motor("BLxxI-MO-TABLE-21:X"),
                1: motor.Motor("BLxxI-MO-TABLE-21:Y"),
                2: motor.Motor("BLxxI-MO-TABLE-21:Z"),
            }
        )
        super().__init__(name)


@pytest.mark.xfail(reason="Flaky test")
@pytest.mark.parametrize("parallel", (False, True))
async def test_many_individual_device_connects_not_slow(parallel):
    start = time.monotonic()
    bundles = [MotorBundle(f"bundle{i}") for i in range(100)]
    if parallel:
        for bundle in bundles:
            await bundle.connect(mock=True)
    else:
        coros = {bundle.name: bundle.connect(mock=True) for bundle in bundles}
        await wait_for_connection(**coros)
    duration = time.monotonic() - start
    assert duration < 1


async def test_device_with_children_lazily_connects(RE):
    parent_motor = MotorBundle("parentMotor")

    for device in [parent_motor, parent_motor.X, parent_motor.Y] + list(
        parent_motor.V.values()
    ):
        assert device._mock is None
    RE(ensure_connected(parent_motor, mock=True))

    for device in [parent_motor, parent_motor.X, parent_motor.Y] + list(
        parent_motor.V.values()
    ):
        assert device._mock is not None


async def test_no_reconnect_signals_if_not_forced():
    parent = DummyDeviceGroup("parent")

    async def inner_connect(mock=False, timeout=None, force_reconnect=False):
        parent.child1.connected = True

    parent.child1.connect = Mock(side_effect=inner_connect)
    await parent.connect(mock=False, timeout=0.01)
    assert parent.child1.connected
    assert parent.child1.connect.call_count == 1
    await parent.connect(mock=False, timeout=0.01)
    assert parent.child1.connected
    assert parent.child1.connect.call_count == 1

    for count in range(2, 10):
        await parent.connect(mock=False, timeout=0.01, force_reconnect=True)
        assert parent.child1.connected
        assert parent.child1.connect.call_count == count


def test_setitem_with_non_int_key():
    device_vector = DeviceVector(children={})
    with pytest.raises(TypeError, match="Expected int, got"):
        device_vector["not_an_int"] = MagicMock(spec=Device)  # type: ignore


def test_setitem_with_non_device_value():
    device_vector = DeviceVector(children={})
    with pytest.raises(TypeError, match="Expected Device, got"):
        device_vector[1] = "not_a_device"



from typing import Annotated as A

from ophyd_async.core import SignalRW, StrictEnum
from ophyd_async.epics import adcore
from ophyd_async.epics.core import PvSuffix


class KinetixTriggerMode(StrictEnum):
    """Trigger mode for ADKinetix detector."""

    INTERNAL = "Internal"
    EDGE = "Rising Edge"
    GATE = "Exp. Gate"


class KinetixReadoutMode(StrictEnum):
    """Readout mode for ADKinetix detector."""

    SENSITIVITY = "1"
    SPEED = "2"
    DYNAMIC_RANGE = "3"
    SUB_ELECTRON = "4"


class KinetixDriverIO(adcore.ADBaseIO):
    """Mirrors the interface provided by ADKinetix/db/ADKinetix.template."""

    trigger_mode: A[SignalRW[KinetixTriggerMode], PvSuffix("TriggerMode")]
    readout_port_idx: A[SignalRW[KinetixReadoutMode], PvSuffix("ReadoutPortIdx")]



import asyncio

import pytest
from bluesky import plan_stubs as bps
from bluesky.run_engine import RunEngine, TransitionError

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    Device,
    NotConnectedError,
    init_devices,
)
from ophyd_async.epics import motor
from ophyd_async.testing import set_mock_value


class FailingDevice(Device):
    async def connect(
        self, mock: bool = False, timeout=DEFAULT_TIMEOUT, force_reconnect=False
    ):
        raise AttributeError()


class WorkingDevice(Device):
    connected = False

    async def connect(
        self, mock: bool = True, timeout=DEFAULT_TIMEOUT, force_reconnect=False
    ):
        self.connected = True
        return await super().connect(mock=True)

    async def set(self, new_position: float): ...


async def test_init_devices_handles_top_level_errors(caplog):
    caplog.set_level(10)
    with pytest.raises(NotConnectedError) as exc:
        async with init_devices():
            _ = FailingDevice("somename")

    assert not exc.value.__cause__

    logs = caplog.get_records("call")
    device_log = [
        log
        for log in logs
        if log.message == "device `_` raised unexpected exception AttributeError"
    ]  # In some environments the asyncio teardown will be logged as an error too

    assert len(device_log) == 1
    assert device_log[0].levelname == "ERROR"


def test_sync_init_devices_no_run_engine_raises_error():
    with pytest.raises(NotConnectedError) as exc:
        with init_devices():
            working_device = WorkingDevice("somename")
    assert exc.value._errors == (
        "Could not connect devices. Is the bluesky event loop running? See "
        "https://blueskyproject.io/ophyd-async/main/"
        "user/explanations/event-loop-choice.html for more info."
    )
    assert not working_device.connected


def test_sync_init_devices_run_engine_created_connects(RE):
    with init_devices():
        working_device = WorkingDevice("somename")

    assert working_device.connected


async def test_init_devices_detects_redeclared_devices():
    original_working_device = working_device = WorkingDevice()

    async with init_devices():
        working_device = WorkingDevice()
    assert original_working_device is not working_device
    assert working_device.connected and working_device.name == "working_device"
    assert not original_working_device.connected and original_working_device.name == ""


def test_connecting_in_plan_raises(RE):
    def bad_plan():
        yield from bps.null()
        with init_devices():
            working_device = WorkingDevice("somename")  # noqa: F841

    with pytest.raises(RuntimeError, match="Cannot use DeviceConnector inside a plan"):
        RE(bad_plan())


def test_async_init_devices_run_engine_same_event_loop():
    async def set_up_device():
        async with init_devices(mock=True):
            mock_motor = motor.Motor("BLxxI-MO-TABLE-01:X")
        set_mock_value(mock_motor.velocity, 1)
        set_mock_value(mock_motor.low_limit_travel, -10.01)
        set_mock_value(mock_motor.high_limit_travel, 20.01)
        return mock_motor

    loop = asyncio.new_event_loop()
    checking_loop = asyncio.new_event_loop()

    try:
        mock_motor = loop.run_until_complete(set_up_device())
        RE = RunEngine(call_returns_result=True, loop=loop)

        def my_plan():
            yield from bps.mov(mock_motor, 3.14)

        RE(my_plan())

        assert (
            checking_loop.run_until_complete(mock_motor.user_setpoint.read())[
                "mock_motor-user_setpoint"
            ]["value"]
            == 3.14
        )

    finally:
        if RE.state not in ("idle", "panicked"):
            try:
                RE.halt()
            except TransitionError:
                pass
        loop.call_soon_threadsafe(loop.stop)
        checking_loop.call_soon_threadsafe(checking_loop.stop)
        RE._th.join()
        loop.close()
        checking_loop.close()


@pytest.mark.skip(
    reason=(
        "MockSignalBackend currently allows a different event-"
        "loop to set the value, unlike real signals."
    )
)
def test_async_init_devices_run_engine_different_event_loop():
    async def set_up_device():
        async with init_devices(mock=True):
            mock_motor = motor.Motor("BLxxI-MO-TABLE-01:X")
        return mock_motor

    init_devices_loop = asyncio.new_event_loop()
    run_engine_loop = asyncio.new_event_loop()
    assert run_engine_loop is not init_devices_loop

    mock_motor = init_devices_loop.run_until_complete(set_up_device())

    RE = RunEngine(loop=run_engine_loop)

    def my_plan():
        yield from bps.mov(mock_motor, 3.14)

    RE(my_plan())

    # The set should fail since the run engine is on a different event loop
    assert (
        init_devices_loop.run_until_complete(mock_motor.user_setpoint.read())[
            "mock_motor-user_setpoint"
        ]["value"]
        != 3.14
    )



import os
import pickle
import socket
import subprocess
import sys
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from random import choice
from typing import Any, Generic, TypeVar

import numpy as np
import pytest
from tango.asyncio_executor import set_global_executor

from ophyd_async.core import Array1D
from ophyd_async.tango.core import DevStateEnum
from ophyd_async.tango.testing import ExampleStrEnum
from ophyd_async.testing import (
    float_array_value,
    int_array_value,
)

T = TypeVar("T")


@pytest.fixture(autouse=True)
def reset_tango_asyncio():
    set_global_executor(None)


def pytest_collection_modifyitems(config, items):
    tango_dir = os.path.join("system_tests", "tango")
    for item in items:
        if tango_dir in str(item.fspath):
            if sys.version_info >= (3, 12):
                item.add_marker(
                    pytest.mark.skip(
                        reason="Tango is currently not supported on Python 3.12: https://github.com/bluesky/ophyd-async/issues/681"
                    )
                )
            if sys.platform.startswith(
                "win"
            ):  # expect "win32", but open to a future change: https://mail.python.org/pipermail/patches/2000-May/000648.html
                item.add_marker(
                    pytest.mark.skip(
                        reason="Ophyd-async is currently not tested on Windows + Tango"
                    )
                )


class TangoSubprocessHelper:
    def __init__(self, args):
        self._args = args

    def __enter__(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.bind(("", 0))
        port = str(self.sock.getsockname()[1])
        self.sock.listen(1)
        subprocess_path = str(Path(__file__).parent / "context_subprocess.py")
        self.process = subprocess.Popen([sys.executable, subprocess_path, port])
        self.conn, _ = self.sock.accept()
        self.conn.send(pickle.dumps(self._args))
        self.trls = pickle.loads(self.conn.recv(1024))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.close()
        self.sock.close()
        self.process.communicate()


@pytest.fixture(scope="module")
def subprocess_helper():
    return TangoSubprocessHelper


@dataclass
class AttributeData(Generic[T]):
    name: str
    py_type: type
    initial: T
    random_put_values: tuple[T, ...]
    cmd_name: str | None

    def random_value(self):
        return choice(self.random_put_values)


class ArrayData(AttributeData):
    def random_value(self):
        array = self.initial.copy()
        for idx in np.ndindex(array.shape):
            array[idx] = choice(self.random_put_values)
        return array


class SequenceData(AttributeData):
    def random_value(self):
        return [choice(self.random_put_values) for _ in range(len(self.initial))]


@pytest.fixture(scope="module")
def everything_signal_info():
    signal_info = {}

    def add_ads(
        name: str,
        tango_type: str,
        py_type: type,
        initial_scalar,
        initial_spectrum,
        choices,
    ):
        scalar_cmd = f"{name}_cmd" if tango_type != "DevUChar" else None
        signal_info[name] = AttributeData(
            name, py_type, initial_scalar, choices, scalar_cmd
        )
        spectrum_cmd = (
            f"{name}_spectrum_cmd"
            if tango_type not in ["DevUChar", "DevState", "DevEnum"]
            else None
        )
        signal_info[f"{name}_spectrum"] = ArrayData(
            f"{name}_spectrum",
            Array1D[py_type],
            initial_spectrum,
            choices,
            spectrum_cmd,
        )
        signal_info[f"{name}_image"] = ArrayData(
            f"{name}_image",
            np.ndarray[Any, np.dtype[py_type]],
            np.vstack((initial_spectrum, initial_spectrum)),
            choices,
            None,
        )

    signal_info["str"] = AttributeData(
        "str", str, "test_string", ("four", "five", "six"), None
    )
    signal_info["str_spectrum"] = SequenceData(
        "str_spectrum",
        Sequence[str],
        ("one", "two", "three"),
        ("four", "five", "six"),
        None,
    )
    signal_info["strenum"] = AttributeData(
        name="strenum",
        py_type=ExampleStrEnum,
        initial=ExampleStrEnum.B,
        random_put_values=[
            ExampleStrEnum.A.value,
            ExampleStrEnum.B.value,
            ExampleStrEnum.C.value,
        ],
        cmd_name=None,
    )
    signal_info["strenum_spectrum"] = SequenceData(
        name="strenum_spectrum",
        py_type=Sequence[ExampleStrEnum],
        initial=[
            ExampleStrEnum.A.value,
            ExampleStrEnum.B.value,
            ExampleStrEnum.C.value,
        ],
        random_put_values=[
            ExampleStrEnum.A.value,
            ExampleStrEnum.B.value,
            ExampleStrEnum.C.value,
        ],
        cmd_name=None,
    )
    add_ads(
        "bool",
        "DevBoolean",
        bool,
        True,
        np.array([False, True], dtype=bool),
        (False, True),
    )
    add_ads("int8", "DevShort", int, 1, int_array_value(np.int8), (1, 2, 3, 4, 5))
    add_ads("uint8", "DevUChar", int, 1, int_array_value(np.uint8), (1, 2, 3, 4, 5))
    add_ads("int16", "DevShort", int, 1, int_array_value(np.int16), (1, 2, 3, 4, 5))
    add_ads("uint16", "DevUShort", int, 1, int_array_value(np.uint16), (1, 2, 3, 4, 5))
    add_ads("int32", "DevLong", int, 1, int_array_value(np.int32), (1, 2, 3, 4, 5))
    add_ads("uint32", "DevULong", int, 1, int_array_value(np.uint32), (1, 2, 3, 4, 5))
    add_ads("int64", "DevLong64", int, 1, int_array_value(np.int64), (1, 2, 3, 4, 5))
    add_ads("uint64", "DevULong64", int, 1, int_array_value(np.uint64), (1, 2, 3, 4, 5))
    add_ads(
        "float32",
        "DevFloat",
        float,
        1.234,
        float_array_value(np.float32),
        (1.234, 2.345, 3.456),
    )
    add_ads(
        "float64",
        "DevDouble",
        float,
        1.234,
        float_array_value(np.float64),
        (1.234, 2.345, 3.456),
    )
    signal_info["my_state"] = AttributeData(
        "my_state",
        DevStateEnum,
        DevStateEnum.INIT,
        random_put_values=[e.name for e in DevStateEnum],
        cmd_name=None,
    )

    signal_info["my_state_spectrum"] = SequenceData(
        "my_state_spectrum",
        Sequence[DevStateEnum],
        initial=[e.name for e in DevStateEnum],
        random_put_values=[e.name for e in DevStateEnum],
        cmd_name=None,
    )

    return signal_info



"""Support for EPICS motor record.

https://github.com/epics-modules/motor
"""

import asyncio

from bluesky.protocols import (
    Flyable,
    Locatable,
    Location,
    Preparable,
    Reading,
    Stoppable,
    Subscribable,
)

from ophyd_async.core import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    AsyncStatus,
    CalculatableTimeout,
    Callback,
    FlyMotorInfo,
    StandardReadable,
    StrictEnum,
    WatchableAsyncStatus,
    WatcherUpdate,
    error_if_none,
    observe_value,
)
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw, epics_signal_w

__all__ = ["MotorLimitsError", "Motor"]


class MotorLimitsError(Exception):
    """Exception for invalid motor limits."""

    pass


# Back compat - delete before 1.0
def __getattr__(name):
    import warnings

    renames = {
        "MotorLimitsException": MotorLimitsError,
    }
    rename = renames.get(name)
    if rename is not None:
        warnings.warn(
            DeprecationWarning(
                f"{name!r} is deprecated, use {rename.__name__!r} instead"
            ),
            stacklevel=2,
        )
        return rename
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


class OffsetMode(StrictEnum):
    VARIABLE = "Variable"
    FROZEN = "Frozen"


class UseSetMode(StrictEnum):
    USE = "Use"
    SET = "Set"


class Motor(
    StandardReadable,
    Locatable[float],
    Stoppable,
    Flyable,
    Preparable,
    Subscribable[float],
):
    """Device that moves a motor record."""

    def __init__(self, prefix: str, name="") -> None:
        # Define some signals
        with self.add_children_as_readables(Format.CONFIG_SIGNAL):
            self.motor_egu = epics_signal_r(str, prefix + ".EGU")
            self.velocity = epics_signal_rw(float, prefix + ".VELO")
            self.offset = epics_signal_rw(float, prefix + ".OFF")

        with self.add_children_as_readables(Format.HINTED_SIGNAL):
            self.user_readback = epics_signal_r(float, prefix + ".RBV")

        self.user_setpoint = epics_signal_rw(float, prefix + ".VAL")
        self.max_velocity = epics_signal_r(float, prefix + ".VMAX")
        self.acceleration_time = epics_signal_rw(float, prefix + ".ACCL")
        self.precision = epics_signal_r(int, prefix + ".PREC")
        self.deadband = epics_signal_r(float, prefix + ".RDBD")
        self.motor_done_move = epics_signal_r(int, prefix + ".DMOV")
        self.low_limit_travel = epics_signal_rw(float, prefix + ".LLM")
        self.high_limit_travel = epics_signal_rw(float, prefix + ".HLM")
        self.offset_freeze_switch = epics_signal_rw(OffsetMode, prefix + ".FOFF")
        self.high_limit_switch = epics_signal_r(int, prefix + ".HLS")
        self.low_limit_switch = epics_signal_r(int, prefix + ".LLS")
        self.set_use_switch = epics_signal_rw(UseSetMode, prefix + ".SET")

        # Note:cannot use epics_signal_x here, as the motor record specifies that
        # we must write 1 to stop the motor. Simply processing the record is not
        # sufficient.
        self.motor_stop = epics_signal_w(int, prefix + ".STOP")

        # Whether set() should complete successfully or not
        self._set_success = True

        # Currently requested fly info, stored in prepare
        self._fly_info: FlyMotorInfo | None = None

        # Set on kickoff(), complete when motor reaches self._fly_completed_position
        self._fly_status: WatchableAsyncStatus | None = None

        super().__init__(name=name)

    def set_name(self, name: str, *, child_name_separator: str | None = None) -> None:
        """Set name of the motor and its children."""
        super().set_name(name, child_name_separator=child_name_separator)
        # Readback should be named the same as its parent in read()
        self.user_readback.set_name(name)

    async def check_motor_limit(self, abs_start_pos: float, abs_end_pos: float):
        """Check the positions are within limits.

        Will raise a MotorLimitsException if the given absolute positions will be
        outside the motor soft limits.
        """
        motor_lower_limit, motor_upper_limit, egu = await asyncio.gather(
            self.low_limit_travel.get_value(),
            self.high_limit_travel.get_value(),
            self.motor_egu.get_value(),
        )

        # EPICS motor record treats limits of 0, 0 as no limit
        if motor_lower_limit == 0 and motor_upper_limit == 0:
            return

        if (
            not motor_upper_limit >= abs_start_pos >= motor_lower_limit
            or not motor_upper_limit >= abs_end_pos >= motor_lower_limit
        ):
            raise MotorLimitsError(
                f"{self.name} motor trajectory for requested fly/move is from "
                f"{abs_start_pos}{egu} to "
                f"{abs_end_pos}{egu} but motor limits are "
                f"{motor_lower_limit}{egu} <= x <= {motor_upper_limit}{egu} "
            )

    @AsyncStatus.wrap
    async def prepare(self, value: FlyMotorInfo):
        """Move to the beginning of a suitable run-up distance ready for a fly scan."""
        self._fly_info = value

        # Velocity, at which motor travels from start_position to end_position, in motor
        # egu/s.
        max_speed, egu = await asyncio.gather(
            self.max_velocity.get_value(), self.motor_egu.get_value()
        )
        if abs(value.velocity) > max_speed:
            raise MotorLimitsError(
                f"Velocity {abs(value.velocity)} {egu}/s was requested for a motor "
                f" with max speed of {max_speed} {egu}/s"
            )

        acceleration_time = await self.acceleration_time.get_value()
        ramp_up_start_pos = value.ramp_up_start_pos(acceleration_time)
        ramp_down_end_pos = value.ramp_down_end_pos(acceleration_time)

        await self.check_motor_limit(ramp_up_start_pos, ramp_down_end_pos)

        # move to prepare position at maximum velocity
        await self.velocity.set(abs(max_speed))
        await self.set(ramp_up_start_pos)

        # Set velocity we will be using for the fly scan
        await self.velocity.set(abs(value.velocity))

    @AsyncStatus.wrap
    async def kickoff(self):
        """Begin moving motor from prepared position to final position."""
        fly_info = error_if_none(
            self._fly_info, "Motor must be prepared before attempting to kickoff"
        )

        acceleration_time = await self.acceleration_time.get_value()
        self._fly_status = self.set(
            fly_info.ramp_down_end_pos(acceleration_time),
            timeout=fly_info.timeout,
        )

    def complete(self) -> WatchableAsyncStatus:
        """Mark as complete once motor reaches completed position."""
        fly_status = error_if_none(self._fly_status, "kickoff not called")
        return fly_status

    @WatchableAsyncStatus.wrap
    async def set(  # type: ignore
        self, new_position: float, timeout: CalculatableTimeout = CALCULATE_TIMEOUT
    ):
        """Move motor to the given value."""
        self._set_success = True
        (
            old_position,
            units,
            precision,
            velocity,
            acceleration_time,
        ) = await asyncio.gather(
            self.user_setpoint.get_value(),
            self.motor_egu.get_value(),
            self.precision.get_value(),
            self.velocity.get_value(),
            self.acceleration_time.get_value(),
        )

        if timeout is CALCULATE_TIMEOUT:
            try:
                timeout = (
                    abs((new_position - old_position) / velocity)
                    + 2 * acceleration_time
                    + DEFAULT_TIMEOUT
                )
            except ZeroDivisionError as error:
                msg = "Mover has zero velocity"
                raise ValueError(msg) from error

        await self.check_motor_limit(old_position, new_position)

        move_status = self.user_setpoint.set(new_position, wait=True, timeout=timeout)
        async for current_position in observe_value(
            self.user_readback, done_status=move_status
        ):
            yield WatcherUpdate(
                current=current_position,
                initial=old_position,
                target=new_position,
                name=self.name,
                unit=units,
                precision=precision,
            )
        if not self._set_success:
            raise RuntimeError("Motor was stopped")

    async def stop(self, success=False):
        """Request to stop moving and return immediately."""
        self._set_success = success
        # Put with completion will never complete as we are waiting for completion on
        # the move above, so need to pass wait=False
        await self.motor_stop.set(1, wait=False)

    async def locate(self) -> Location[float]:
        """Return the current setpoint and readback of the motor."""
        setpoint, readback = await asyncio.gather(
            self.user_setpoint.get_value(), self.user_readback.get_value()
        )
        return Location(setpoint=setpoint, readback=readback)

    def subscribe_reading(self, function: Callback[dict[str, Reading[float]]]) -> None:
        """Subscribe to reading."""
        self.user_readback.subscribe_reading(function)

    subscribe = subscribe_reading

    def clear_sub(self, function: Callback[dict[str, Reading[float]]]) -> None:
        """Unsubscribe."""
        self.user_readback.clear_sub(function)
